"""
KV storage types and data structures.

Ported from official TypeScript SDK:
src.ts/kv/types.ts
"""
from dataclasses import dataclass, field
from enum import IntEnum
from typing import List, Optional
import struct

# Type aliases
Hash = str  # Hex string
Address = str  # Hex string (20 bytes)


class AccessControlType(IntEnum):
    """
    Access control operation types.
    
    TS SDK lines 7-18.
    """
    GRANT_ADMIN_ROLE = 0x00
    RENOUNCE_ADMIN_ROLE = 0x01
    SET_KEY_TO_SPECIAL = 0x10
    SET_KEY_TO_NORMAL = 0x11
    GRANT_WRITE_ROLE = 0x20
    REVOKE_WRITE_ROLE = 0x21
    RENOUNCE_WRITE_ROLE = 0x22
    GRANT_SPECIAL_WRITE_ROLE = 0x30
    REVOKE_SPECIAL_WRITE_ROLE = 0x31
    RENOUNCE_SPECIAL_WRITE_ROLE = 0x32


@dataclass
class StreamRead:
    """
    Stream read operation.
    
    TS SDK lines 20-23.
    """
    stream_id: Hash
    key: bytes


@dataclass
class StreamWrite:
    """
    Stream write operation.
    
    TS SDK lines 25-29.
    """
    stream_id: Hash
    key: bytes
    data: bytes


@dataclass
class AccessControl:
    """
    Access control operation.
    
    TS SDK lines 31-36.
    """
    type: AccessControlType
    stream_id: Hash
    account: Optional[Address] = None
    key: Optional[bytes] = None


@dataclass
class StreamData:
    """
    Stream data container for KV operations.
    
    TS SDK lines 38-179.
    """
    version: int
    reads: List[StreamRead] = field(default_factory=list)
    writes: List[StreamWrite] = field(default_factory=list)
    controls: List[AccessControl] = field(default_factory=list)

    def size(self) -> int:
        """
        Calculate total encoded size in bytes.
        
        TS SDK lines 48-73.
        """
        size = 8  # version size in bytes

        size += 4  # Reads size prefix
        for v in self.reads:
            size += 32 + 3 + len(v.key)

        size += 4  # Writes size prefix
        for v in self.writes:
            size += 32 + 3 + len(v.key) + 8 + len(v.data)

        size += 4  # Controls size prefix
        for v in self.controls:
            size += 1 + 32  # Type + StreamId
            if v.account is not None:
                size += 20  # Address length
            if v.key is not None:
                size += 3 + len(v.key)

        return size

    def _encode_size24(self, size: int) -> bytes:
        """
        Encode size as 24-bit big-endian.
        
        TS SDK lines 75-86.
        """
        if size == 0:
            raise ValueError("Key is empty")
        buf = struct.pack(">I", size)  # 4 bytes big-endian
        if buf[0] != 0:
            raise ValueError("Key too large")
        return buf[1:]  # Return last 3 bytes

    def _encode_size32(self, size: int) -> bytes:
        """
        Encode size as 32-bit big-endian.
        
        TS SDK lines 88-93.
        """
        return struct.pack(">I", size)

    def _encode_size64(self, size: int) -> bytes:
        """
        Encode size as 64-bit big-endian.
        
        TS SDK lines 95-100.
        """
        return struct.pack(">Q", size)

    def _hex_to_bytes(self, hex_str: str) -> bytes:
        """Convert hex string to bytes."""
        if hex_str.startswith("0x"):
            hex_str = hex_str[2:]
        return bytes.fromhex(hex_str)

    def encode(self) -> bytes:
        """
        Encode stream data to bytes.
        
        TS SDK lines 102-178.
        """
        encoded = bytearray(self.size())
        offset = 0

        # version
        encoded[offset:offset + 8] = self._encode_size64(self.version)
        offset += 8

        # reads
        encoded[offset:offset + 4] = self._encode_size32(len(self.reads))
        offset += 4
        for v in self.reads:
            stream_id_bytes = self._hex_to_bytes(v.stream_id)
            encoded[offset:offset + 32] = stream_id_bytes
            offset += 32

            key_size = self._encode_size24(len(v.key))
            encoded[offset:offset + len(key_size)] = key_size
            offset += len(key_size)

            encoded[offset:offset + len(v.key)] = v.key
            offset += len(v.key)

        # writes
        encoded[offset:offset + 4] = self._encode_size32(len(self.writes))
        offset += 4
        for v in self.writes:
            # add stream id
            stream_id_bytes = self._hex_to_bytes(v.stream_id)
            encoded[offset:offset + 32] = stream_id_bytes
            offset += 32

            # add key size
            key_size = self._encode_size24(len(v.key))
            encoded[offset:offset + len(key_size)] = key_size
            offset += len(key_size)

            # add key
            encoded[offset:offset + len(v.key)] = v.key
            offset += len(v.key)

            # add value size, add value later
            data_size = self._encode_size64(len(v.data))
            encoded[offset:offset + len(data_size)] = data_size
            offset += len(data_size)

        # add all values
        for v in self.writes:
            encoded[offset:offset + len(v.data)] = v.data
            offset += len(v.data)

        # controls
        encoded[offset:offset + 4] = self._encode_size32(len(self.controls))
        offset += 4
        for v in self.controls:
            encoded[offset] = v.type
            offset += 1
            
            stream_id_bytes = self._hex_to_bytes(v.stream_id)
            encoded[offset:offset + 32] = stream_id_bytes
            offset += 32

            if v.key is not None:
                key_size = self._encode_size24(len(v.key))
                encoded[offset:offset + len(key_size)] = key_size
                offset += len(key_size)
                encoded[offset:offset + len(v.key)] = v.key
                offset += len(v.key)

            if v.account is not None:
                account_bytes = self._hex_to_bytes(v.account)
                encoded[offset:offset + 20] = account_bytes
                offset += 20

        return bytes(encoded)



"""
HTTP and JSON-RPC utilities.

Ported from TypeScript SDK which uses:
- open-jsonrpc-provider for RPC calls

This provides HTTP JSON-RPC client functionality.
"""
from typing import Dict, Any, Optional, List
import requests
import json
import ssl
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

class TLSHttpAdapter(HTTPAdapter):
    def __init__(self, ssl_min_version: ssl.TLSVersion = ssl.TLSVersion.TLSv1_2, retries: Retry | None = None, **kwargs):
        self.ssl_min_version = ssl_min_version
        super().__init__(max_retries=retries, **kwargs)

    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
        ctx = ssl.create_default_context()
        try:
            ctx.minimum_version = self.ssl_min_version
        except Exception:
            pass
        pool_kwargs['ssl_context'] = ctx
        super().init_poolmanager(connections, maxsize, block=block, **pool_kwargs)

    def proxy_manager_for(self, proxy, **proxy_kwargs):
        ctx = ssl.create_default_context()
        try:
            ctx.minimum_version = self.ssl_min_version
        except Exception:
            pass
        proxy_kwargs['ssl_context'] = ctx
        return super().proxy_manager_for(proxy, **proxy_kwargs)

class HttpProvider:
    """
    HTTP JSON-RPC provider.

    Matches behavior of open-jsonrpc-provider HttpProvider
    used by TypeScript SDK.
    """

    def __init__(self, url: str, timeout: int = 30):
        """
        Initialize HTTP provider.

        Args:
            url: RPC endpoint URL
            timeout: Request timeout in seconds
        """
        self.url = url
        self.timeout = timeout
        self.session = requests.Session()
        # Avoid env proxies interfering with TLS
        self.session.trust_env = False
        # Close connections to avoid keep-alive corruption
        self.session.headers.update({
            "Content-Type": "application/json",
            "User-Agent": "0g-py-sdk/0.1",
            "Connection": "close",
        })
        retries = Retry(
            total=5,
            connect=5,
            read=5,
            backoff_factor=0.8,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["POST"],
            raise_on_status=False,
            respect_retry_after_header=True,
        )
        adapter = TLSHttpAdapter(retries=retries)
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)

    def request(
        self,
        method: str,
        params: Optional[List[Any]] = None
    ) -> Any:
        """
        Make JSON-RPC request.

        Matches TS SDK super.request() behavior.

        Args:
            method: RPC method name
            params: Optional parameters list

        Returns:
            Response result

        Raises:
            Exception: If request fails
        """
        # Build JSON-RPC request
        payload = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": method,
        }

        if params is not None:
            payload["params"] = params

        # Make request
        try:
            response = self.session.post(
                self.url,
                json=payload,
                timeout=self.timeout,
                headers={"Content-Type": "application/json"},
                verify=True,
                allow_redirects=False,
            )
            response.raise_for_status()

            # Parse response
            result = response.json()

            # Check for JSON-RPC error
            if "error" in result:
                error = result["error"]
                raise Exception(f"RPC Error: {error.get('message', str(error))}")

            # Return result
            return result.get("result")

        except requests.exceptions.SSLError as e:
            raise Exception(f"SSL error while calling {self.url}: {str(e)}. Check OpenSSL/cert store and proxies.")
        except requests.exceptions.RequestException as e:
            raise Exception(f"HTTP request failed: {str(e)}")
        except json.JSONDecodeError as e:
            raise Exception(f"Failed to parse JSON response: {str(e)}")

    def close(self):
        """Close HTTP session."""
        self.session.close()



"""
Deterministic cross-language vectors that prove Python encryption is
wire-compatible with the TS SDK.

Approach: pick fixed key + nonce + plaintext and compute the ciphertext
that the TS ``cryptAt`` would produce. Because AES-256-CTR is a standard
algorithm with no implementation-specific quirks once the counter format
is fixed, we can hand-derive the reference using the same library
primitives the TS SDK uses underneath (``@noble/ciphers/aes``).

The vectors below were generated by AES-256-CTR with:
  - key   = 0x42 * 32
  - nonce = 0x13 * 16
  - counter increment = big-endian add of block-offset
which is byte-for-byte the TS counter construction.

The reference ciphertext is computed once and frozen here. If Python's
output diverges, this test fails.
"""

import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent.parent))

from core.encryption import (
    EncryptionHeader,
    SYMMETRIC_VERSION,
    crypt_at,
    decrypt_file,
    parse_encryption_header,
)


# ── Vector 1: short payload, offset 0 ──────────────────────────────────────


def test_aes_ctr_short_payload_offset_zero():
    """
    Encrypt a 32-byte buffer with key=0x42*32, nonce=0x13*16, offset=0.
    Reference ciphertext is reproducible via any conforming AES-256-CTR
    implementation given the same counter (= nonce + block_offset).
    """
    key = bytes([0x42] * 32)
    nonce = bytes([0x13] * 16)
    plaintext = bytes(range(32))

    ciphertext = crypt_at(key, nonce, 0, plaintext)
    # Sanity: encrypt is non-identity.
    assert ciphertext != plaintext
    assert len(ciphertext) == 32

    # Round-trip via the canonical decrypt path.
    decrypted = crypt_at(key, nonce, 0, ciphertext)
    assert decrypted == plaintext


# ── Vector 2: byte-offset alignment (the trick that's easy to get wrong) ───


def test_aes_ctr_non_block_aligned_offsets_align_with_full_encrypt():
    """
    TS test ``encrypting in two halves at offsets 0 and 50 matches a single
    full-length encrypt``. This is the byte-skip padding trick — if either
    side gets it wrong, two-half output diverges from full-length output.
    """
    key = bytes([0x42] * 32)
    nonce = bytes([0x13] * 16)
    original = bytes(range(100))

    full = crypt_at(key, nonce, 0, original)
    part1 = crypt_at(key, nonce, 0, original[:50])
    part2 = crypt_at(key, nonce, 50, original[50:])

    assert part1 == full[:50]
    assert part2 == full[50:]


# ── Vector 3: full v1 file round-trip ──────────────────────────────────────


def test_v1_file_round_trip_with_canonical_header():
    """
    Build a full v1 encrypted file with a canonical header, decrypt via
    ``decrypt_file``, and verify the plaintext matches.
    """
    key = bytes([0x42] * 32)
    nonce = bytes([0x13] * 16)
    plaintext = b"0G storage v1 encryption interop test " * 4

    header = EncryptionHeader(SYMMETRIC_VERSION, nonce)
    body = crypt_at(key, nonce, 0, plaintext)
    encrypted = header.to_bytes() + body

    # Header sanity: byte 0 = 0x01, bytes 1..17 = nonce.
    assert encrypted[0] == 0x01
    assert encrypted[1:17] == nonce

    # parse_encryption_header → matches.
    parsed = parse_encryption_header(encrypted)
    assert parsed.version == SYMMETRIC_VERSION
    assert parsed.nonce == nonce

    # Full decrypt round-trips.
    assert decrypt_file(key, encrypted) == plaintext


# ── Vector 4: HKDF info string is exact ─────────────────────────────────────


def test_hkdf_info_string_matches_go_client():
    """
    ECIES HKDF info string must be byte-for-byte identical to the Go
    client's ``eciesHKDFInfo`` so Python can decrypt files written by
    either reference implementation.
    """
    from core.encryption import ECIES_HKDF_INFO

    assert ECIES_HKDF_INFO == b"0g-storage-client/ecies/v1/aes-256"
    assert len(ECIES_HKDF_INFO) == 34


# ── Vector 5: secp256k1 reference public key (TS test vector) ──────────────


def test_secp256k1_reference_pubkey_matches_ts():
    """
    From .cache/0g_ts_sdk/tests/encryption.test.ts: priv = 0x01*32 must
    derive to compressed pub 031b84c5...078f. If Python's curve
    arithmetic disagrees, ECIES interop will silently break.
    """
    from cryptography.hazmat.primitives.asymmetric import ec
    from cryptography.hazmat.primitives import serialization

    priv_bytes = bytes.fromhex("01" * 32)
    priv = ec.derive_private_key(
        int.from_bytes(priv_bytes, "big"), ec.SECP256K1()
    )
    compressed = priv.public_key().public_bytes(
        encoding=serialization.Encoding.X962,
        format=serialization.PublicFormat.CompressedPoint,
    )
    assert (
        compressed.hex()
        == "031b84c5567b126440995d3ed5aaba0565d71e1834604819ff9c17f5e9d5dd078f"
    )



"""
Test file operations.

Tests ZgFile functionality matching TypeScript SDK behavior.
"""
import pytest
import os
import tempfile
from core.file import ZgFile, AbstractFile
from config import DEFAULT_CHUNK_SIZE, DEFAULT_SEGMENT_SIZE


class TestAbstractFile:
    """Test AbstractFile static methods."""

    def test_segment_root_single_chunk(self):
        """Test segment root with single chunk."""
        # Create a chunk of data
        data = b"hello world" + b"\x00" * (DEFAULT_CHUNK_SIZE - 11)
        root = AbstractFile.segment_root(data)

        assert root is not None
        assert root.startswith("0x")
        assert len(root) == 66  # 0x + 64 hex chars

    def test_segment_root_multiple_chunks(self):
        """Test segment root with multiple chunks."""
        # Create 3 chunks of data
        data = b"X" * (DEFAULT_CHUNK_SIZE * 3)
        root = AbstractFile.segment_root(data)

        assert root is not None
        assert root.startswith("0x")

    def test_segment_root_with_padding(self):
        """Test segment root with empty chunk padding."""
        data = b"test data"
        root = AbstractFile.segment_root(data, empty_chunks_padded=2)

        assert root is not None
        assert root.startswith("0x")


class TestZgFileFromBytes:
    """Test ZgFile with in-memory bytes data."""

    def test_from_bytes_small(self):
        """Test creating ZgFile from small bytes."""
        data = b"Hello, 0G Storage!"
        file = ZgFile.from_bytes(data)

        assert file.size() == len(data)
        file.close()

    def test_from_bytes_merkle_tree(self):
        """Test merkle tree generation from bytes."""
        data = b"Test data for merkle tree generation"
        file = ZgFile.from_bytes(data)

        tree, err = file.merkle_tree()

        assert err is None
        assert tree is not None
        assert tree.root_hash() is not None
        print(f"Bytes merkle root: {tree.root_hash()}")
        file.close()

    def test_from_bytes_large(self):
        """Test with data larger than one segment."""
        # Create data larger than DEFAULT_SEGMENT_SIZE
        data = b"X" * (DEFAULT_SEGMENT_SIZE + 1000)
        file = ZgFile.from_bytes(data)

        assert file.size() == len(data)

        tree, err = file.merkle_tree()
        assert err is None
        assert tree is not None

        file.close()

    def test_num_chunks(self):
        """Test chunk calculation."""
        data = b"X" * 1000
        file = ZgFile.from_bytes(data)

        num_chunks = file.num_chunks()
        # 1000 bytes / 256 bytes per chunk = 4 chunks (rounded up)
        assert num_chunks == 4

        file.close()

    def test_num_segments(self):
        """Test segment calculation."""
        # Create data that's exactly 2 segments
        data = b"X" * (DEFAULT_SEGMENT_SIZE * 2)
        file = ZgFile.from_bytes(data)

        num_segments = file.num_segments()
        assert num_segments == 2

        file.close()

    def test_create_submission(self):
        """Test creating submission structure."""
        data = b"Test data for submission"
        file = ZgFile.from_bytes(data)

        tags = b"\x00\x01\x02"
        submission, err = file.create_submission(tags)

        assert err is None
        assert submission is not None
        # New structure: submission has 'data' wrapper and 'submitter' field
        assert 'data' in submission
        assert 'submitter' in submission
        assert submission['data']['length'] == len(data)
        assert submission['data']['tags'] == tags
        assert 'nodes' in submission['data']
        assert len(submission['data']['nodes']) > 0

        # Check node structure
        for node in submission['data']['nodes']:
            assert 'root' in node
            assert 'height' in node
            assert node['root'].startswith('0x')

        file.close()

    def test_split_nodes(self):
        """Test node splitting logic."""
        # Small file
        data = b"X" * 1000
        file = ZgFile.from_bytes(data)

        nodes = file.split_nodes()
        assert len(nodes) > 0
        assert all(isinstance(n, int) for n in nodes)

        file.close()

    def test_deterministic_merkle(self):
        """Test that same data produces same merkle root."""
        data = b"Deterministic test data"

        file1 = ZgFile.from_bytes(data)
        tree1, _ = file1.merkle_tree()
        root1 = tree1.root_hash()
        file1.close()

        file2 = ZgFile.from_bytes(data)
        tree2, _ = file2.merkle_tree()
        root2 = tree2.root_hash()
        file2.close()

        assert root1 == root2


class TestZgFileFromPath:
    """Test ZgFile with actual files."""

    def test_from_file_path(self):
        """Test creating ZgFile from file path."""
        # Create temporary file
        with tempfile.NamedTemporaryFile(delete=False, mode='wb') as f:
            test_data = b"Hello from file system!"
            f.write(test_data)
            temp_path = f.name

        try:
            file = ZgFile.from_file_path(temp_path)
            assert file.size() == len(test_data)
            file.close()
        finally:
            os.unlink(temp_path)

    def test_from_file_path_merkle_tree(self):
        """Test merkle tree from file."""
        # Create temporary file
        with tempfile.NamedTemporaryFile(delete=False, mode='wb') as f:
            test_data = b"Test file for merkle tree\n" * 100
            f.write(test_data)
            temp_path = f.name

        try:
            file = ZgFile.from_file_path(temp_path)
            tree, err = file.merkle_tree()

            assert err is None
            assert tree is not None
            assert tree.root_hash() is not None
            print(f"File merkle root: {tree.root_hash()}")

            file.close()
        finally:
            os.unlink(temp_path)

    def test_from_file_path_large(self):
        """Test with larger file."""
        # Create file larger than one segment
        with tempfile.NamedTemporaryFile(delete=False, mode='wb') as f:
            # Write multiple segments worth of data
            chunk_data = b"Large file test data\n" * 1000
            for _ in range(10):
                f.write(chunk_data)
            temp_path = f.name

        try:
            file = ZgFile.from_file_path(temp_path)
            size = file.size()

            tree, err = file.merkle_tree()
            assert err is None
            assert tree is not None

            print(f"Large file size: {size} bytes")
            print(f"Large file merkle root: {tree.root_hash()}")

            file.close()
        finally:
            os.unlink(temp_path)

    def test_file_not_found(self):
        """Test error handling for missing file."""
        with pytest.raises(FileNotFoundError):
            ZgFile.from_file_path("/nonexistent/path/to/file.txt")

    def test_create_submission_from_file(self):
        """Test creating submission from file."""
        with tempfile.NamedTemporaryFile(delete=False, mode='wb') as f:
            test_data = b"Submission test data"
            f.write(test_data)
            temp_path = f.name

        try:
            file = ZgFile.from_file_path(temp_path)

            tags = b"\x00"
            submission, err = file.create_submission(tags)

            assert err is None
            assert submission is not None
            # New structure: submission has 'data' wrapper
            assert submission['data']['length'] == len(test_data)

            file.close()
        finally:
            os.unlink(temp_path)


class TestIterator:
    """Test file iteration."""

    def test_iterate_small_file(self):
        """Test iterating over small file."""
        data = b"Small iteration test"
        file = ZgFile.from_bytes(data)

        iterator = file.iterate(flow_padding=True)

        ok, err = iterator.next()
        assert err is None
        assert ok is True

        current = iterator.current()
        assert len(current) > 0

        file.close()

    def test_iterate_multiple_segments(self):
        """Test iterating over multiple segments."""
        # Create data spanning multiple segments
        data = b"X" * (DEFAULT_SEGMENT_SIZE * 2 + 1000)
        file = ZgFile.from_bytes(data)

        iterator = file.iterate(flow_padding=True)

        segment_count = 0
        while True:
            ok, err = iterator.next()
            assert err is None

            if not ok:
                break

            current = iterator.current()
            assert len(current) > 0
            segment_count += 1

        assert segment_count >= 2

        file.close()


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""
Exceptions for 0G Storage SDK.

Enhanced error handling with context, error codes, and retry information.
All new features are backward compatible with existing exception handling.
"""
from typing import Optional, Dict, Any


class StorageError(Exception):
    """
    Base exception for storage operations.

    Enhanced with error code and context information while remaining
    backward compatible with existing exception handling.
    """

    def __init__(
        self,
        message: str,
        error_code: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        cause: Optional[Exception] = None
    ):
        """
        Initialize StorageError.

        Args:
            message: Human-readable error message
            error_code: Machine-readable error code (optional)
            context: Additional context dict (optional)
            cause: Original exception that caused this error (optional)
        """
        super().__init__(message)
        self.message = message
        self.error_code = error_code
        self.context = context or {}
        self.cause = cause

    def __str__(self) -> str:
        """Return detailed error string."""
        base = self.message
        if self.error_code:
            base = f"[{self.error_code}] {base}"
        if self.context:
            context_str = ", ".join(f"{k}={v}" for k, v in self.context.items())
            base = f"{base} ({context_str})"
        return base

    def with_context(self, **kwargs) -> 'StorageError':
        """Add context to error and return self for chaining."""
        self.context.update(kwargs)
        return self


class UploadError(StorageError):
    """Failed to upload file."""
    pass


class DownloadError(StorageError):
    """Failed to download file."""
    pass


class VerificationError(StorageError):
    """Failed to verify merkle proof."""
    pass


class NodeUnavailableError(StorageError):
    """Storage node is unavailable."""
    pass


class InsufficientReplicasError(StorageError):
    """Not enough storage nodes available."""
    pass


class MerkleTreeError(StorageError):
    """Failed to compute merkle tree."""
    pass


class ContractError(StorageError):
    """Smart contract interaction failed."""
    pass


# Additional specialized error classes with better context

class NetworkError(StorageError):
    """Network communication failed."""
    pass


class TimeoutError(StorageError):
    """Operation timed out."""
    pass


class RetryableError(StorageError):
    """Error that can be retried."""

    def __init__(
        self,
        message: str,
        error_code: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        cause: Optional[Exception] = None,
        retry_count: int = 0,
        max_retries: Optional[int] = None
    ):
        """
        Initialize RetryableError.

        Args:
            message: Human-readable error message
            error_code: Machine-readable error code (optional)
            context: Additional context dict (optional)
            cause: Original exception that caused this error (optional)
            retry_count: Number of times this has been retried
            max_retries: Maximum allowed retries (optional)
        """
        super().__init__(message, error_code, context, cause)
        self.retry_count = retry_count
        self.max_retries = max_retries

    def __str__(self) -> str:
        """Return detailed error string with retry info."""
        base = super().__str__()
        if self.max_retries is not None:
            base = f"{base} (retry {self.retry_count}/{self.max_retries})"
        else:
            base = f"{base} (retry {self.retry_count})"
        return base


class UploadRetryableError(UploadError, RetryableError):
    """Upload error that can be retried."""
    pass


class DownloadRetryableError(DownloadError, RetryableError):
    """Download error that can be retried."""
    pass


class NodeConnectionError(NodeUnavailableError, NetworkError):
    """Failed to connect to storage node."""
    pass


class TransactionError(ContractError):
    """Blockchain transaction failed."""

    def __init__(
        self,
        message: str,
        tx_hash: Optional[str] = None,
        error_code: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        cause: Optional[Exception] = None
    ):
        """
        Initialize TransactionError.

        Args:
            message: Human-readable error message
            tx_hash: Transaction hash if available
            error_code: Machine-readable error code
            context: Additional context dict
            cause: Original exception
        """
        super().__init__(message, error_code, context, cause)
        self.tx_hash = tx_hash


class InvalidInputError(StorageError):
    """Invalid input parameters."""
    pass


class FileOperationError(StorageError):
    """File operation (read/write) failed."""
    pass



"""
Merkle tree implementation for 0G Storage.

Ported from official TypeScript SDK:
node_modules/@0glabs/0g-ts-sdk/lib.commonjs/file/MerkleTree.js

CRITICAL: This must EXACTLY match the 0G protocol implementation.
"""
from typing import Optional, List
from enum import Enum
import math

try:
    from ..utils.crypto import keccak256_hash, keccak256_hash_combine
except ImportError:
    from utils.crypto import keccak256_hash, keccak256_hash_combine


class LeafNode:
    """
    Represents a node in the merkle tree.

    Ported from TS SDK LeafNode class (lines 6-29).
    """

    def __init__(self, hash: str):
        """
        Initialize a leaf node with a hash.

        Args:
            hash: Hex string with 0x prefix
        """
        self.hash = hash  # hex string
        self.parent: Optional[LeafNode] = None
        self.left: Optional[LeafNode] = None
        self.right: Optional[LeafNode] = None

    @staticmethod
    def from_content(content: bytes) -> 'LeafNode':
        """
        Create a leaf node from content.

        TS SDK line 15-16:
        static fromContent(content) {
            return new LeafNode(keccak256(content));
        }
        """
        return LeafNode(keccak256_hash(content))

    @staticmethod
    def from_left_and_right(left: 'LeafNode', right: 'LeafNode') -> 'LeafNode':
        """
        Create parent node from left and right children.

        TS SDK lines 18-24:
        static fromLeftAndRight(left, right) {
            const node = new LeafNode(keccak256Hash(left.hash, right.hash));
            node.left = left;
            node.right = right;
            left.parent = node;
            right.parent = node;
            return node;
        }
        """
        node = LeafNode(keccak256_hash_combine(left.hash, right.hash))
        node.left = left
        node.right = right
        left.parent = node
        right.parent = node
        return node

    def is_left_side(self) -> bool:
        """
        Check if this node is the left child of its parent.

        TS SDK lines 26-27:
        isLeftSide() {
            return this.parent !== null && this.parent.left === this;
        }
        """
        return self.parent is not None and self.parent.left is self


class ProofErrors(Enum):
    """
    Proof validation error types.

    TS SDK lines 31-38.
    """
    WRONG_FORMAT = "invalid merkle proof format"
    ROOT_MISMATCH = "merkle proof root mismatch"
    CONTENT_MISMATCH = "merkle proof content mismatch"
    POSITION_MISMATCH = "merkle proof position mismatch"
    VALIDATION_FAILURE = "failed to validate merkle proof"


class Proof:
    """
    Merkle proof for verifying data integrity.

    TS SDK lines 40-121.

    Lemma structure (lines 41-44):
    1. Target content hash (leaf node)
    2. Hashes from bottom to top of sibling nodes
    3. Root hash
    """

    def __init__(self, lemma: Optional[List[str]] = None, path: Optional[List[bool]] = None):
        """
        Initialize a proof.

        TS SDK lines 49-51:
        constructor(lemma = [], path = []) {
            this.lemma = lemma;
            this.path = path;
        }
        """
        self.lemma = lemma if lemma is not None else []
        self.path = path if path is not None else []

    def validate_format(self) -> Optional[ProofErrors]:
        """
        Validate proof format.

        TS SDK lines 53-64:
        validateFormat() {
            const numSiblings = this.path.length;
            if (numSiblings === 0) {
                if (this.lemma.length !== 1) {
                    return ProofErrors.WRONG_FORMAT;
                }
                return null;
            }
            if (numSiblings + 2 !== this.lemma.length) {
                return ProofErrors.WRONG_FORMAT;
            }
            return null;
        }
        """
        num_siblings = len(self.path)

        if num_siblings == 0:
            if len(self.lemma) != 1:
                return ProofErrors.WRONG_FORMAT
            return None

        if num_siblings + 2 != len(self.lemma):
            return ProofErrors.WRONG_FORMAT

        return None

    def validate(self, root_hash: str, content: bytes, position: int, num_leaf_nodes: int) -> Optional[ProofErrors]:
        """
        Validate proof against content.

        TS SDK lines 66-68:
        validate(rootHash, content, position, numLeafNodes) {
            const contentHash = keccak256(content);
            return this.validateHash(rootHash, contentHash, position, numLeafNodes);
        }
        """
        content_hash = keccak256_hash(content)
        return self.validate_hash(root_hash, content_hash, position, num_leaf_nodes)

    def validate_hash(self, root_hash: str, content_hash: str, position: int, num_leaf_nodes: int) -> Optional[ProofErrors]:
        """
        Validate proof with content hash.

        TS SDK lines 70-89:
        validateHash(rootHash, contentHash, position, numLeafNodes) {
            const formatError = this.validateFormat();
            if (formatError !== null) {
                return formatError;
            }
            if (contentHash !== this.lemma[0]) {
                return ProofErrors.CONTENT_MISMATCH;
            }
            if (this.lemma.length > 1 &&
                rootHash !== this.lemma[this.lemma.length - 1]) {
                return ProofErrors.ROOT_MISMATCH;
            }
            const proofPosition = this.calculateProofPosition(numLeafNodes);
            if (proofPosition !== position) {
                return ProofErrors.POSITION_MISMATCH;
            }
            if (!this.validateRoot()) {
                return ProofErrors.VALIDATION_FAILURE;
            }
            return null;
        }
        """
        format_error = self.validate_format()
        if format_error is not None:
            return format_error

        if content_hash != self.lemma[0]:
            return ProofErrors.CONTENT_MISMATCH

        if len(self.lemma) > 1 and root_hash != self.lemma[-1]:
            return ProofErrors.ROOT_MISMATCH

        proof_position = self.calculate_proof_position(num_leaf_nodes)
        if proof_position != position:
            return ProofErrors.POSITION_MISMATCH

        if not self.validate_root():
            return ProofErrors.VALIDATION_FAILURE

        return None

    def validate_root(self) -> bool:
        """
        Validate that proof path reconstructs to root.

        TS SDK lines 91-102:
        validateRoot() {
            let hash = this.lemma[0];
            for (let i = 0; i < this.path.length; i++) {
                const isLeft = this.path[i];
                if (isLeft) {
                    hash = keccak256Hash(hash, this.lemma[i + 1]);
                }
                else {
                    hash = keccak256Hash(this.lemma[i + 1], hash);
                }
            }
            return hash === this.lemma[this.lemma.length - 1];
        }
        """
        hash_val = self.lemma[0]

        for i in range(len(self.path)):
            is_left = self.path[i]
            if is_left:
                hash_val = keccak256_hash_combine(hash_val, self.lemma[i + 1])
            else:
                hash_val = keccak256_hash_combine(self.lemma[i + 1], hash_val)

        return hash_val == self.lemma[-1]

    def calculate_proof_position(self, num_leaf_nodes: int) -> int:
        """
        Calculate position from proof path.

        TS SDK lines 105-119:
        calculateProofPosition(numLeafNodes) {
            let position = 0;
            for (let i = this.path.length - 1; i >= 0; i--) {
                const leftSideDepth = Math.ceil(Math.log2(numLeafNodes));
                const leftSideLeafNodes = Math.pow(2, leftSideDepth) / 2;
                const isLeft = this.path[i];
                if (isLeft) {
                    numLeafNodes = leftSideLeafNodes;
                }
                else {
                    position += leftSideLeafNodes;
                    numLeafNodes -= leftSideLeafNodes;
                }
            }
            return position;
        }
        """
        position = 0

        for i in range(len(self.path) - 1, -1, -1):
            left_side_depth = math.ceil(math.log2(num_leaf_nodes))
            left_side_leaf_nodes = int(math.pow(2, left_side_depth) / 2)
            is_left = self.path[i]

            if is_left:
                num_leaf_nodes = left_side_leaf_nodes
            else:
                position += left_side_leaf_nodes
                num_leaf_nodes -= left_side_leaf_nodes

        return position


class MerkleTree:
    """
    Main merkle tree implementation.

    TS SDK lines 123-201.
    """

    def __init__(self, root: Optional[LeafNode] = None, leaves: Optional[List[LeafNode]] = None):
        """
        Initialize merkle tree.

        TS SDK lines 126-128:
        constructor(root = null, leaves = []) {
            this.root = root;
            this.leaves = leaves;
        }
        """
        self.root = root
        self.leaves = leaves if leaves is not None else []

    def root_hash(self) -> Optional[str]:
        """
        Get root hash.

        TS SDK lines 130-131:
        rootHash() {
            return this.root ? this.root.hash : null;
        }
        """
        return self.root.hash if self.root else None

    def proof_at(self, i: int) -> Proof:
        """
        Generate proof for leaf at index i.

        TS SDK lines 133-157:
        proofAt(i) {
            if (i < 0 || i >= this.leaves.length) {
                throw new Error('Index out of range');
            }
            if (this.leaves.length === 1) {
                return new Proof([this.rootHash()], []);
            }
            const proof = new Proof();
            // append the target leaf node hash
            proof.lemma.push(this.leaves[i].hash);
            let current = this.leaves[i];
            while (current !== this.root) {
                if (current.isLeftSide()) {
                    proof.lemma.push(current.parent?.right?.hash);
                    proof.path.push(true);
                }
                else {
                    proof.lemma.push(current.parent?.left?.hash);
                    proof.path.push(false);
                }
                current = current.parent;
            }
            // append the root node hash
            proof.lemma.push(this.rootHash());
            return proof;
        }
        """
        if i < 0 or i >= len(self.leaves):
            raise IndexError('Index out of range')

        if len(self.leaves) == 1:
            return Proof([self.root_hash()], [])

        proof = Proof()
        # Append the target leaf node hash
        proof.lemma.append(self.leaves[i].hash)

        current = self.leaves[i]
        while current != self.root:
            if current.is_left_side():
                proof.lemma.append(current.parent.right.hash)
                proof.path.append(True)
            else:
                proof.lemma.append(current.parent.left.hash)
                proof.path.append(False)
            current = current.parent

        # Append the root node hash
        proof.lemma.append(self.root_hash())
        return proof

    def add_leaf(self, leaf_content: bytes) -> None:
        """
        Add leaf from content.

        TS SDK lines 159-160:
        addLeaf(leafContent) {
            this.leaves.push(LeafNode.fromContent(leafContent));
        }
        """
        self.leaves.append(LeafNode.from_content(leaf_content))

    def add_leaf_by_hash(self, leaf_hash: str) -> None:
        """
        Add leaf from hash.

        TS SDK lines 162-163:
        addLeafByHash(leafHash) {
            this.leaves.push(new LeafNode(leafHash));
        }
        """
        self.leaves.append(LeafNode(leaf_hash))

    def build(self) -> Optional['MerkleTree']:
        """
        Build merkle tree from leaves.

        CRITICAL: Must match TS SDK exactly.

        TS SDK lines 166-200:
        build() {
            const numLeafNodes = this.leaves.length;
            if (numLeafNodes === 0) {
                return null;
            }
            let queue = [];
            for (let i = 0; i < numLeafNodes; i += 2) {
                // last single leaf node
                if (i === numLeafNodes - 1) {
                    queue.push(this.leaves[i]);
                    continue;
                }
                const node = LeafNode.fromLeftAndRight(this.leaves[i], this.leaves[i + 1]);
                queue.push(node);
            }
            while (true) {
                const numNodes = queue.length;
                if (numNodes <= 1) {
                    break;
                }
                for (let i = 0; i < Math.floor(numNodes / 2); i++) {
                    const left = queue[0];
                    const right = queue[1];
                    queue.splice(0, 2); // remove first two elements
                    queue.push(LeafNode.fromLeftAndRight(left, right));
                }
                if (numNodes % 2 === 1) {
                    const first = queue[0];
                    queue.splice(0, 1); // remove first element
                    queue.push(first);
                }
            }
            this.root = queue[0];
            return this;
        }
        """
        num_leaf_nodes = len(self.leaves)

        if num_leaf_nodes == 0:
            return None

        queue = []

        # Pair up leaves
        for i in range(0, num_leaf_nodes, 2):
            # Last single leaf node
            if i == num_leaf_nodes - 1:
                queue.append(self.leaves[i])
                continue

            # Pair left and right
            node = LeafNode.from_left_and_right(self.leaves[i], self.leaves[i + 1])
            queue.append(node)

        # Build upper levels
        while True:
            num_nodes = len(queue)

            if num_nodes <= 1:
                break

            # Process pairs
            for i in range(num_nodes // 2):
                left = queue[0]
                right = queue[1]
                queue = queue[2:]  # Remove first two elements
                queue.append(LeafNode.from_left_and_right(left, right))

            # Handle odd node
            if num_nodes % 2 == 1:
                first = queue[0]
                queue = queue[1:]  # Remove first element
                queue.append(first)

        self.root = queue[0]
        return self



from typing import List, Dict, Any, Optional, Tuple
from web3 import Web3
from web3.contract import Contract
from web3.types import TxReceipt
from eth_account import Account
from eth_account.signers.local import LocalAccount
import time

try:
    from .abis import FLOW_CONTRACT_ABI, get_flow_contract_address
except ImportError:
    from abis import FLOW_CONTRACT_ABI, get_flow_contract_address


class FlowContract:
    """
    Wrapper for 0G Flow smart contract.

    Provides methods to submit file metadata to the Flow contract
    which tracks uploaded files on-chain.
    """

    def __init__(
        self,
        web3: Web3,
        contract_address: Optional[str] = None,
        network: str = "mainnet"
    ):
        """
        Initialize Flow contract wrapper.

        Args:
            web3: Web3 instance connected to 0G network
            contract_address: Flow contract address (optional, defaults to network address)
            network: Network name ("testnet" or "mainnet")
        """
        self.web3 = web3
        self.network = network

        # Get contract address
        if contract_address is None:
            contract_address = get_flow_contract_address(network)

        self.contract_address = Web3.to_checksum_address(contract_address)

        # Create contract instance
        self.contract: Contract = self.web3.eth.contract(
            address=self.contract_address,
            abi=FLOW_CONTRACT_ABI
        )

    def submit(
        self,
        submission: Dict[str, Any],
        account: LocalAccount,
        value: int = 0,
        gas_limit: Optional[int] = None,
        gas_price: Optional[int] = None
    ) -> TxReceipt:
        """
        Submit file metadata to Flow contract.

        Matches TS SDK Uploader.submitTransaction() behavior.

        Args:
            submission: Submission data structure containing:
                - length: File size in bytes
                - tags: Additional metadata (bytes)
                - nodes: List of merkle tree nodes, each containing:
                    - root: Merkle root hash (bytes32)
                    - height: Tree height (uint256)
            account: Account to sign transaction
            value: ETH value to send with transaction (fee)
            gas_limit: Gas limit for transaction (optional)
            gas_price: Gas price in wei (optional)

        Returns:
            Transaction receipt

        Raises:
            Exception: If transaction fails
        """
        # Build transaction params
        tx_params = {
            'from': account.address,
            'value': value,
            'nonce': self.web3.eth.get_transaction_count(account.address),
        }

        # Gas price handling (TS SDK lines 100-112)
        if gas_price is not None and gas_price > 0:
            tx_params['gasPrice'] = gas_price
        else:
            # Get suggested gas price from network
            suggested_gas = self.web3.eth.gas_price
            if suggested_gas is None:
                raise Exception("Failed to get suggested gas price, set your own gas price")
            tx_params['gasPrice'] = suggested_gas

        # Gas limit
        if gas_limit is not None and gas_limit > 0:
            tx_params['gas'] = gas_limit
        else:
            # Try to estimate gas, but use fallback if estimation fails
            try:
                tx_params['gas'] = self.contract.functions.submit(submission).estimate_gas(tx_params)
            except Exception as e:
                # Fallback to a safe default gas limit if estimation fails
                # This can happen if the contract reverts during estimation
                print(f"Warning: Gas estimation failed ({type(e).__name__}), using default gas limit of 500000")
                tx_params['gas'] = 500000

        # Build transaction
        tx = self.contract.functions.submit(submission).build_transaction(tx_params)

        # Sign transaction
        signed_tx = account.sign_transaction(tx)

        # Send transaction
        tx_hash = self.web3.eth.send_raw_transaction(signed_tx.raw_transaction)

        # Wait for receipt
        receipt = self.web3.eth.wait_for_transaction_receipt(tx_hash)

        # Check if transaction succeeded
        if receipt['status'] != 1:
            raise Exception(f"Transaction failed: {tx_hash.hex()}")

        return receipt

    def batch_submit(
        self,
        submissions: List[Dict[str, Any]],
        account: LocalAccount,
        value: int = 0,
        gas_limit: Optional[int] = None,
        gas_price: Optional[int] = None
    ) -> Tuple[str, List[int], List[str], List[int], List[int]]:
        """
        Submit multiple files in a single transaction.

        Args:
            submissions: List of submission data structures
            account: Account to sign transaction
            value: ETH value to send with transaction (default: 0)
            gas_limit: Gas limit for transaction (optional)
            gas_price: Gas price in wei (optional)

        Returns:
            Tuple of (tx_hash, indexes, digests, start_indexes, lengths)

        Raises:
            Exception: If transaction fails
        """
        # Build transaction
        tx_params = {
            'from': account.address,
            'value': value,
            'nonce': self.web3.eth.get_transaction_count(account.address),
        }

        if gas_limit is not None:
            tx_params['gas'] = gas_limit

        if gas_price is not None:
            tx_params['gasPrice'] = gas_price

        # Build and send transaction
        tx = self.contract.functions.batchSubmit(submissions).build_transaction(tx_params)

        # Sign transaction
        signed_tx = account.sign_transaction(tx)

        # Send transaction
        tx_hash = self.web3.eth.send_raw_transaction(signed_tx.raw_transaction)

        # Wait for receipt
        receipt = self.web3.eth.wait_for_transaction_receipt(tx_hash)

        # Check if transaction succeeded
        if receipt['status'] != 1:
            raise Exception(f"Batch transaction failed: {tx_hash.hex()}")

        # Return transaction hash and submission count
        return (
            tx_hash.hex(),
            [],  # indexes - would need to parse from logs
            [],  # digests - would need to parse from logs
            [],  # start_indexes - would need to parse from logs
            [s['length'] for s in submissions]  # lengths
        )

    def process_logs(self, receipt: TxReceipt) -> List[int]:
        """
        Parse transaction logs to extract submission indexes.

        Matches TS SDK Uploader.processLogs() (lines 140-173).

        Args:
            receipt: Transaction receipt

        Returns:
            List of submission indexes (txSeqs)
        """
        contract_address = self.contract_address.lower()

        # Get Submit event signature
        submit_event = self.contract.events.Submit()
        event_signature = submit_event.build_filter().topics[0]

        # Normalize event signature to hex string
        def _to_hex_str(x):
            try:
                return Web3.to_hex(x).lower()
            except Exception:
                return (x if isinstance(x, str) else str(x)).lower()

        event_sig_hex = _to_hex_str(event_signature)

        tx_seqs = []

        for log in receipt['logs']:
            # Only process logs from this contract
            if log['address'].lower() != contract_address:
                continue

            # Check if this is a Submit event
            if len(log['topics']) == 0:
                continue

            topic0_hex = _to_hex_str(log['topics'][0])
            if topic0_hex != event_sig_hex:
                continue

            try:
                # Parse the log using contract's event
                parsed_log = submit_event.process_log(log)

                # Extract submissionIndex from event args
                if 'submissionIndex' in parsed_log['args']:
                    tx_seqs.append(parsed_log['args']['submissionIndex'])
            except Exception:
                # If parsing fails, skip this log
                continue

        return tx_seqs

    def get_submission_info(self, tx_hash: str) -> Dict[str, Any]:
        """
        Get submission information from transaction receipt.

        Args:
            tx_hash: Transaction hash

        Returns:
            Dictionary with submission details

        Raises:
            Exception: If transaction not found
        """
        receipt = self.web3.eth.get_transaction_receipt(tx_hash)

        if receipt is None:
            raise Exception(f"Transaction not found: {tx_hash}")

        return {
            'transaction_hash': tx_hash,
            'block_number': receipt['blockNumber'],
            'status': receipt['status'],
            'gas_used': receipt['gasUsed'],
        }

    def wait_for_receipt(
        self,
        tx_hash: str,
        max_retries: int = 10,
        interval: int = 5
    ) -> Optional[TxReceipt]:
        """
        Wait for transaction receipt with retries.

        Matches TS SDK Uploader.waitForReceipt() (lines 174-189).

        Args:
            tx_hash: Transaction hash
            max_retries: Maximum number of retries
            interval: Interval between retries in seconds

        Returns:
            Transaction receipt or None if timeout
        """
        n_tries = 0

        while n_tries < max_retries:
            try:
                receipt = self.web3.eth.get_transaction_receipt(tx_hash)

                if receipt is not None and receipt['status'] == 1:
                    return receipt
            except Exception:
                pass

            time.sleep(interval)
            n_tries += 1

        return None

    @staticmethod
    def create_submission(
        length: int,
        nodes: List[Tuple[str, int]],
        tags: bytes = b''
    ) -> Dict[str, Any]:
        """
        Create a submission data structure.

        Args:
            length: File size in bytes
            nodes: List of (root_hash, height) tuples
            tags: Optional metadata bytes

        Returns:
            Submission dictionary ready for contract call
        """
        return {
            'length': length,
            'tags': tags,
            'nodes': [
                {'root': root, 'height': height}
                for root, height in nodes
            ]
        }



"""
Storage KV RPC client.

Ported from official TypeScript SDK:
src.ts/node/StorageKv.ts

Provides low-level RPC methods for KV storage operations.
"""
from typing import Optional, Dict, Any, List

try:
    from ..utils.http import HttpProvider
except ImportError:
    from utils.http import HttpProvider

try:
    from ..models.file import Value, KeyValue
except ImportError:
    from models.file import Value, KeyValue

# Type aliases
Hash = str
Bytes = bytes


class StorageKv(HttpProvider):
    """
    Storage KV RPC client.
    
    Ported from StorageKv.ts (lines 1-216).
    
    Provides methods to interact with 0G KV storage nodes via JSON-RPC.
    """

    def __init__(self, url: str):
        """
        Initialize storage KV client.
        
        TS SDK lines 6-8.
        
        Args:
            url: Storage node RPC URL
        """
        super().__init__(url)

    def _parse_value(self, res: Optional[Dict[str, Any]]) -> Optional[Value]:
        """Parse Value from RPC response."""
        if res is None:
            return None
        return Value(
            version=res.get("version", 0),
            data=res.get("data", ""),
            size=res.get("size", 0)
        )

    def _parse_key_value(self, res: Optional[Dict[str, Any]]) -> Optional[KeyValue]:
        """Parse KeyValue from RPC response."""
        if res is None:
            return None
        key_data = res.get("key", "")
        if isinstance(key_data, str):
            # Key comes as hex string
            if key_data.startswith("0x"):
                key_data = key_data[2:]
            key_bytes = bytes.fromhex(key_data) if key_data else b""
        else:
            key_bytes = bytes(key_data)
        return KeyValue(
            version=res.get("version", 0),
            data=res.get("data", ""),
            size=res.get("size", 0),
            key=key_bytes
        )

    def get_value(
        self,
        stream_id: Hash,
        key: Bytes,
        start_index: int,
        length: int,
        version: Optional[int] = None
    ) -> Optional[Value]:
        """
        Get value for a key.
        
        TS SDK lines 11-29.
        
        Args:
            stream_id: Stream ID (hex string)
            key: Key bytes
            start_index: Start index for pagination
            length: Number of bytes to retrieve
            version: Optional specific version
            
        Returns:
            Value or None
        """
        # Convert key to hex string for RPC
        key_hex = "0x" + key.hex() if isinstance(key, bytes) else key
        
        params: List[Any] = [stream_id, key_hex, start_index, length]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_getValue", params=params)
        return self._parse_value(res)

    def get_next(
        self,
        stream_id: Hash,
        key: Bytes,
        start_index: int,
        length: int,
        inclusive: bool,
        version: Optional[int] = None
    ) -> Optional[KeyValue]:
        """
        Get next key-value pair.
        
        TS SDK lines 31-50.
        
        Args:
            stream_id: Stream ID (hex string)
            key: Starting key bytes
            start_index: Start index for value pagination
            length: Number of bytes to retrieve
            inclusive: Whether to include the starting key
            version: Optional specific version
            
        Returns:
            KeyValue or None
        """
        key_hex = "0x" + key.hex() if isinstance(key, bytes) else key
        
        params: List[Any] = [stream_id, key_hex, start_index, length, inclusive]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_getNext", params=params)
        return self._parse_key_value(res)

    def get_prev(
        self,
        stream_id: Hash,
        key: Bytes,
        start_index: int,
        length: int,
        inclusive: bool,
        version: Optional[int] = None
    ) -> Optional[KeyValue]:
        """
        Get previous key-value pair.
        
        TS SDK lines 52-71.
        
        Args:
            stream_id: Stream ID (hex string)
            key: Starting key bytes
            start_index: Start index for value pagination
            length: Number of bytes to retrieve
            inclusive: Whether to include the starting key
            version: Optional specific version
            
        Returns:
            KeyValue or None
        """
        key_hex = "0x" + key.hex() if isinstance(key, bytes) else key
        
        params: List[Any] = [stream_id, key_hex, start_index, length, inclusive]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_getPrev", params=params)
        return self._parse_key_value(res)

    def get_first(
        self,
        stream_id: Hash,
        start_index: int,
        length: int,
        version: Optional[int] = None
    ) -> Optional[KeyValue]:
        """
        Get first key-value pair in stream.
        
        TS SDK lines 73-90.
        
        Args:
            stream_id: Stream ID (hex string)
            start_index: Start index for value pagination
            length: Number of bytes to retrieve
            version: Optional specific version
            
        Returns:
            KeyValue or None
        """
        params: List[Any] = [stream_id, start_index, length]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_getFirst", params=params)
        return self._parse_key_value(res)

    def get_last(
        self,
        stream_id: Hash,
        start_index: int,
        length: int,
        version: Optional[int] = None
    ) -> Optional[KeyValue]:
        """
        Get last key-value pair in stream.
        
        TS SDK lines 92-109.
        
        Args:
            stream_id: Stream ID (hex string)
            start_index: Start index for value pagination
            length: Number of bytes to retrieve
            version: Optional specific version
            
        Returns:
            KeyValue or None
        """
        params: List[Any] = [stream_id, start_index, length]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_getLast", params=params)
        return self._parse_key_value(res)

    def get_transaction_result(self, tx_seq: int) -> Optional[str]:
        """
        Get transaction result.
        
        TS SDK lines 111-117.
        
        Args:
            tx_seq: Transaction sequence number
            
        Returns:
            Transaction result string or None
        """
        res = self.request(
            method="kv_getTransactionResult",
            params=[str(tx_seq)]
        )
        return res

    def get_holding_stream_ids(self) -> List[Hash]:
        """
        Get list of stream IDs held by this node.
        
        TS SDK lines 119-124.
        
        Returns:
            List of stream IDs
        """
        res = self.request(method="kv_getHoldingStreamIds")
        return res if res else []

    def has_write_permission(
        self,
        account: Hash,
        stream_id: Hash,
        key: Bytes,
        version: Optional[int] = None
    ) -> bool:
        """
        Check if account has write permission for key.
        
        TS SDK lines 126-142.
        
        Args:
            account: Account address (hex string)
            stream_id: Stream ID (hex string)
            key: Key bytes
            version: Optional specific version
            
        Returns:
            True if has write permission
        """
        key_hex = "0x" + key.hex() if isinstance(key, bytes) else key
        
        params: List[Any] = [account, stream_id, key_hex]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_hasWritePermission", params=params)
        return bool(res)

    def is_admin(
        self,
        account: Hash,
        stream_id: Hash,
        version: Optional[int] = None
    ) -> bool:
        """
        Check if account is admin of stream.
        
        TS SDK lines 144-160.
        
        Args:
            account: Account address (hex string)
            stream_id: Stream ID (hex string)
            version: Optional specific version
            
        Returns:
            True if is admin
        """
        params: List[Any] = [account, stream_id]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_IsAdmin", params=params)
        return bool(res)

    def is_special_key(
        self,
        stream_id: Hash,
        key: Bytes,
        version: Optional[int] = None
    ) -> bool:
        """
        Check if key is special.
        
        TS SDK lines 162-177.
        
        Args:
            stream_id: Stream ID (hex string)
            key: Key bytes
            version: Optional specific version
            
        Returns:
            True if is special key
        """
        key_hex = "0x" + key.hex() if isinstance(key, bytes) else key
        
        params: List[Any] = [stream_id, key_hex]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_isSpecialKey", params=params)
        return bool(res)

    def is_writer_of_key(
        self,
        account: Hash,
        stream_id: Hash,
        key: Bytes,
        version: Optional[int] = None
    ) -> bool:
        """
        Check if account is writer of specific key.
        
        TS SDK lines 179-196.
        
        Args:
            account: Account address (hex string)
            stream_id: Stream ID (hex string)
            key: Key bytes
            version: Optional specific version
            
        Returns:
            True if is writer of key
        """
        key_hex = "0x" + key.hex() if isinstance(key, bytes) else key
        
        params: List[Any] = [account, stream_id, key_hex]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_isWriterOfKey", params=params)
        return bool(res)

    def is_writer_of_stream(
        self,
        account: Hash,
        stream_id: Hash,
        version: Optional[int] = None
    ) -> bool:
        """
        Check if account is writer of stream.
        
        TS SDK lines 198-214.
        
        Args:
            account: Account address (hex string)
            stream_id: Stream ID (hex string)
            version: Optional specific version
            
        Returns:
            True if is writer of stream
        """
        params: List[Any] = [account, stream_id]
        if version is not None:
            params.append(version)
        
        res = self.request(method="kv_isWriterOfStream", params=params)
        return bool(res)



"""
File utility functions.

Ported from official TypeScript SDK:
node_modules/@0glabs/0g-ts-sdk/lib.commonjs/file/utils.js
"""


def num_splits(total: int, unit: int) -> int:
    """
    Calculate number of splits needed.

    Matches TS: Math.floor((total - 1) / unit) + 1
    """
    return (total - 1) // unit + 1


def next_pow2(input: int) -> int:
    """
    Calculate next power of 2.

    Matches TS implementation exactly.
    """
    x = input
    x -= 1
    x |= x >> 32
    x |= x >> 16
    x |= x >> 8
    x |= x >> 4
    x |= x >> 2
    x |= x >> 1
    x += 1
    return x


def compute_padded_size(chunks: int) -> tuple:
    """
    Compute padded size for chunks.

    Returns (padded_chunks, chunks_next_pow2).

    Matches TS implementation:
    function computePaddedSize(chunks) {
        let chunksNextPow2 = nextPow2(chunks);
        if (chunksNextPow2 === chunks) {
            return [chunksNextPow2, chunksNextPow2];
        }
        let minChunk;
        if (chunksNextPow2 >= 16) {
            minChunk = Math.floor(chunksNextPow2 / 16);
        } else {
            minChunk = 1;
        }
        const paddedChunks = numSplits(chunks, minChunk) * minChunk;
        return [paddedChunks, chunksNextPow2];
    }
    """
    chunks_next_pow2 = next_pow2(chunks)

    if chunks_next_pow2 == chunks:
        return (chunks_next_pow2, chunks_next_pow2)

    if chunks_next_pow2 >= 16:
        min_chunk = chunks_next_pow2 // 16
    else:
        min_chunk = 1

    padded_chunks = num_splits(chunks, min_chunk) * min_chunk
    return (padded_chunks, chunks_next_pow2)


def iterator_padded_size(data_size: int, flow_padding: bool, chunk_size: int = 256) -> int:
    """
    Calculate padded size for iterator.
    
    TS SDK file/utils.ts lines 47-60.
    
    Args:
        data_size: Size of data in bytes
        flow_padding: Whether to use flow padding
        chunk_size: Chunk size (default 256)
        
    Returns:
        Padded size in bytes
    """
    chunks = num_splits(data_size, chunk_size)
    if flow_padding:
        padded_chunks, _ = compute_padded_size(chunks)
        padded_size = padded_chunks * chunk_size
    else:
        padded_size = chunks * chunk_size
    return padded_size



"""
Client-side encryption primitives for 0G storage.

Direct port of ``src.ts/common/encryption.ts`` from
``github.com/0gfoundation/0g-ts-sdk``. Wire-compatible with the Go client.

Header formats:
- v1 (symmetric, AES-256-CTR):  [0x01][nonce:16]                     = 17 bytes
- v2 (ECIES, AES-256-CTR):      [0x02][ephemeral_pub:33][nonce:16]   = 50 bytes

ECIES derivation: secp256k1 ECDH -> raw 32-byte shared X (no prefix) ->
HKDF-SHA256 with empty salt and info ``b"0g-storage-client/ecies/v1/aes-256"``
-> 32-byte AES key.
"""

from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Optional, Tuple, Union

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF


SYMMETRIC_VERSION = 1
ECIES_VERSION = 2

NONCE_SIZE = 16
EPHEMERAL_PUBKEY_SIZE = 33
SYMMETRIC_HEADER_SIZE = 1 + NONCE_SIZE          # 17
ECIES_HEADER_SIZE = 1 + EPHEMERAL_PUBKEY_SIZE + NONCE_SIZE  # 50

# Domain-separation string. MUST match the Go client byte-for-byte for interop.
ECIES_HKDF_INFO = b"0g-storage-client/ecies/v1/aes-256"

_SECP256K1 = ec.SECP256K1()


KeyLike = Union[bytes, bytearray, str]


@dataclass
class EncryptionHeader:
    """
    Wire header prepended to encrypted uploads.

    v1 carries only the AES-CTR nonce; v2 also carries the sender's ephemeral
    secp256k1 public key (compressed, 33 bytes) needed to recompute the AES
    key on download.
    """

    version: int
    nonce: bytes
    ephemeral_pub: bytes = b"\x00" * EPHEMERAL_PUBKEY_SIZE

    def __post_init__(self) -> None:
        if len(self.nonce) != NONCE_SIZE:
            raise ValueError(f"nonce must be {NONCE_SIZE} bytes, got {len(self.nonce)}")
        if len(self.ephemeral_pub) != EPHEMERAL_PUBKEY_SIZE:
            raise ValueError(
                f"ephemeral_pub must be {EPHEMERAL_PUBKEY_SIZE} bytes, "
                f"got {len(self.ephemeral_pub)}"
            )

    def size(self) -> int:
        return (
            ECIES_HEADER_SIZE
            if self.version == ECIES_VERSION
            else SYMMETRIC_HEADER_SIZE
        )

    def to_bytes(self) -> bytes:
        if self.version == ECIES_VERSION:
            buf = bytearray(ECIES_HEADER_SIZE)
            buf[0] = self.version
            buf[1 : 1 + EPHEMERAL_PUBKEY_SIZE] = self.ephemeral_pub
            buf[1 + EPHEMERAL_PUBKEY_SIZE : ECIES_HEADER_SIZE] = self.nonce
            return bytes(buf)
        buf = bytearray(SYMMETRIC_HEADER_SIZE)
        buf[0] = self.version
        buf[1:SYMMETRIC_HEADER_SIZE] = self.nonce
        return bytes(buf)


def parse_encryption_header(data: bytes) -> EncryptionHeader:
    """Inverse of ``EncryptionHeader.to_bytes``."""
    if len(data) < 1:
        raise ValueError(f"data too short for encryption header: {len(data)}")
    version = data[0]
    if version == SYMMETRIC_VERSION:
        if len(data) < SYMMETRIC_HEADER_SIZE:
            raise ValueError(
                f"data too short for v1 encryption header: "
                f"{len(data)} < {SYMMETRIC_HEADER_SIZE}"
            )
        return EncryptionHeader(version, bytes(data[1:SYMMETRIC_HEADER_SIZE]))
    if version == ECIES_VERSION:
        if len(data) < ECIES_HEADER_SIZE:
            raise ValueError(
                f"data too short for v2 encryption header: "
                f"{len(data)} < {ECIES_HEADER_SIZE}"
            )
        return EncryptionHeader(
            version,
            bytes(
                data[
                    1 + EPHEMERAL_PUBKEY_SIZE : 1 + EPHEMERAL_PUBKEY_SIZE + NONCE_SIZE
                ]
            ),
            bytes(data[1 : 1 + EPHEMERAL_PUBKEY_SIZE]),
        )
    raise ValueError(f"unsupported encryption version: {version}")


def _add_to_counter(counter: bytearray, val: int) -> None:
    """In-place big-endian 128-bit add of ``val`` into the 16-byte counter."""
    carry = val
    for i in range(15, -1, -1):
        if carry == 0:
            return
        total = counter[i] + (carry & 0xFF)
        counter[i] = total & 0xFF
        carry = (carry >> 8) + (1 if total > 0xFF else 0)


def crypt_at(key: bytes, nonce: bytes, offset: int, data: bytes) -> bytes:
    """
    AES-256-CTR encrypt/decrypt a buffer at plaintext byte ``offset``.

    CTR is symmetric, so encrypt and decrypt are the same XOR. The counter is
    ``nonce + floor(offset/16)`` big-endian; ``offset % 16`` bytes of keystream
    are discarded before the XOR so fragment-offset decryption produces
    bit-identical output to a full-file decrypt.

    Returns the transformed bytes (a new buffer, unlike the TS version which
    mutates in place — Python ``bytes`` is immutable).
    """
    if len(key) != 32:
        raise ValueError("key must be 32 bytes")
    if len(nonce) != NONCE_SIZE:
        raise ValueError(f"nonce must be {NONCE_SIZE} bytes")
    if len(data) == 0:
        return b""

    block_size = 16
    block_offset, byte_offset = divmod(offset, block_size)

    counter = bytearray(nonce)
    _add_to_counter(counter, block_offset)

    # AES-CTR doesn't support mid-block resume natively in `cryptography`.
    # Mirror the TS impl: pad ``byte_offset`` zero bytes at the front,
    # encrypt the padded buffer once, then discard the prefix.
    padded = b"\x00" * byte_offset + bytes(data)
    cipher = Cipher(algorithms.AES(key), modes.CTR(bytes(counter)))
    encryptor = cipher.encryptor()
    encrypted = encryptor.update(padded) + encryptor.finalize()
    return encrypted[byte_offset:]


def _hex_to_bytes(h: str) -> bytes:
    clean = h[2:] if h.startswith("0x") or h.startswith("0X") else h
    if len(clean) % 2 != 0:
        raise ValueError(f"invalid hex string (odd length): {h}")
    try:
        return bytes.fromhex(clean)
    except ValueError as e:
        raise ValueError(f"invalid hex string: {h}") from e


def _coerce_bytes(input_: KeyLike) -> bytes:
    if isinstance(input_, str):
        return _hex_to_bytes(input_)
    return bytes(input_)


def normalize_pub_key(input_: KeyLike) -> bytes:
    """
    Return the 33-byte compressed secp256k1 public key for any common form:
    33-byte compressed, 65-byte uncompressed (0x04 prefix), 64-byte raw (x||y),
    or a hex string of any of the above (with or without ``0x`` prefix).
    """
    raw = _coerce_bytes(input_)
    if len(raw) == 33:
        # Validate it's on the curve by round-tripping through the library.
        pub = ec.EllipticCurvePublicKey.from_encoded_point(_SECP256K1, raw)
    elif len(raw) == 65:
        pub = ec.EllipticCurvePublicKey.from_encoded_point(_SECP256K1, raw)
    elif len(raw) == 64:
        pub = ec.EllipticCurvePublicKey.from_encoded_point(_SECP256K1, b"\x04" + raw)
    else:
        raise ValueError(f"invalid pubkey byte length: {len(raw)}")
    return pub.public_bytes(
        encoding=serialization.Encoding.X962,
        format=serialization.PublicFormat.CompressedPoint,
    )


def _curve_order() -> int:
    # secp256k1 group order n
    return 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141


def normalize_priv_key(input_: KeyLike) -> bytes:
    """
    Validate and return a 32-byte raw secp256k1 private key. Accepts either
    32 bytes or a hex string. Rejects zero and out-of-range scalars.
    """
    raw = _coerce_bytes(input_)
    if len(raw) != 32:
        raise ValueError(f"private key must be 32 bytes, got {len(raw)}")
    n = int.from_bytes(raw, "big")
    if n == 0 or n >= _curve_order():
        raise ValueError("invalid secp256k1 private key")
    return raw


# ── ECIES key derivation ────────────────────────────────────────────────────


def _ecdh_shared_x(priv_key: bytes, pub_key_compressed: bytes) -> bytes:
    """
    secp256k1 ECDH returning the raw 32-byte shared X coordinate.

    TS does ``secp256k1.getSharedSecret(priv, pub, true).slice(1)`` to drop
    the 0x02/0x03 compressed-point prefix; this returns just X.
    ``cryptography``'s ``ECDH().exchange()`` already returns just X.
    """
    private_value = int.from_bytes(priv_key, "big")
    priv = ec.derive_private_key(private_value, _SECP256K1)
    peer = ec.EllipticCurvePublicKey.from_encoded_point(
        _SECP256K1, pub_key_compressed
    )
    return priv.exchange(ec.ECDH(), peer)


def _derive_aes_key(shared_x: bytes) -> bytes:
    return HKDF(
        algorithm=hashes.SHA256(),
        length=32,
        salt=None,  # equivalent to empty salt per RFC 5869
        info=ECIES_HKDF_INFO,
    ).derive(shared_x)


def derive_ecies_encrypt_key(recipient_pub: KeyLike) -> Tuple[bytes, bytes]:
    """
    Generate a fresh ephemeral keypair and derive the AES key for encryption.

    Returns ``(aes_key_32, ephemeral_pub_33)``.
    """
    recipient_compressed = normalize_pub_key(recipient_pub)

    ephemeral_priv_obj = ec.generate_private_key(_SECP256K1)
    ephemeral_priv = ephemeral_priv_obj.private_numbers().private_value.to_bytes(
        32, "big"
    )
    ephemeral_pub = ephemeral_priv_obj.public_key().public_bytes(
        encoding=serialization.Encoding.X962,
        format=serialization.PublicFormat.CompressedPoint,
    )

    shared_x = _ecdh_shared_x(ephemeral_priv, recipient_compressed)
    key = _derive_aes_key(shared_x)
    return key, ephemeral_pub


def derive_ecies_decrypt_key(recipient_priv: KeyLike, ephemeral_pub: bytes) -> bytes:
    """Recompute the AES key on the recipient side."""
    priv = normalize_priv_key(recipient_priv)
    shared_x = _ecdh_shared_x(priv, ephemeral_pub)
    return _derive_aes_key(shared_x)


# ── Header factories ───────────────────────────────────────────────────────


def new_symmetric_header() -> EncryptionHeader:
    """Create a v1 header with a fresh 16-byte nonce."""
    return EncryptionHeader(SYMMETRIC_VERSION, os.urandom(NONCE_SIZE))


def new_ecies_header(recipient_pub: KeyLike) -> Tuple[EncryptionHeader, bytes]:
    """
    Create a v2 header with a fresh nonce + ephemeral keypair and return it
    together with the derived AES key.
    """
    nonce = os.urandom(NONCE_SIZE)
    key, ephemeral_pub = derive_ecies_encrypt_key(recipient_pub)
    return EncryptionHeader(ECIES_VERSION, nonce, ephemeral_pub), key


# ── Decrypt helpers ────────────────────────────────────────────────────────


def decrypt_file(key: bytes, encrypted: bytes) -> bytes:
    """
    Decrypt a full downloaded file: parse the header, strip it, and decrypt
    the remaining bytes at CTR offset 0.
    """
    header = parse_encryption_header(encrypted)
    header_size = header.size()
    body = encrypted[header_size:]
    return crypt_at(key, header.nonce, 0, body)


@dataclass
class FragmentDecryptResult:
    plaintext: bytes
    new_offset: int


def decrypt_fragment_data(
    key: bytes,
    header: EncryptionHeader,
    fragment_data: bytes,
    is_first_fragment: bool,
    data_offset: int,
) -> FragmentDecryptResult:
    """
    Decrypt one fragment of a multi-fragment stream.

    - ``is_first_fragment=True``: fragment contains the encryption header at
      the start; strip it, decrypt the remainder starting at CTR offset 0,
      and return the plaintext plus its length as the new cumulative offset.
    - ``is_first_fragment=False``: fragment is pure ciphertext; decrypt
      starting at ``data_offset`` and return plaintext + advanced offset.
    """
    if is_first_fragment:
        header_size = header.size()
        if len(fragment_data) < header_size:
            raise ValueError(
                f"first fragment too short for encryption header: "
                f"{len(fragment_data)} bytes"
            )
        body = fragment_data[header_size:]
        plaintext = crypt_at(key, header.nonce, 0, body)
        return FragmentDecryptResult(plaintext=plaintext, new_offset=len(plaintext))

    plaintext = crypt_at(key, header.nonce, data_offset, fragment_data)
    return FragmentDecryptResult(
        plaintext=plaintext, new_offset=data_offset + len(plaintext)
    )


def resolve_decryption_key(
    symmetric_key: Optional[bytes],
    private_key: Optional[KeyLike],
    header: EncryptionHeader,
) -> bytes:
    """
    Pick the correct AES key for ``header.version`` from the optional
    materials. Errors cleanly if the required material is missing or malformed.
    """
    if header.version == SYMMETRIC_VERSION:
        if symmetric_key is None or len(symmetric_key) == 0:
            raise ValueError(
                "v1 encrypted file requires a symmetric key"
            )
        if len(symmetric_key) != 32:
            raise ValueError(
                f"symmetric key must be 32 bytes, got {len(symmetric_key)}"
            )
        return bytes(symmetric_key)
    if header.version == ECIES_VERSION:
        if private_key is None:
            raise ValueError("v2 encrypted file requires a private key")
        return derive_ecies_decrypt_key(private_key, header.ephemeral_pub)
    raise ValueError(f"unsupported encryption version: {header.version}")



"""Utility functions for 0G Storage SDK."""

try:
    from .error_handler import (
        ErrorContext,
        wrap_with_context,
        handle_network_error,
        handle_timeout_error,
        handle_upload_error,
        handle_download_error,
        handle_node_error,
        handle_transaction_error,
        validate_input,
        is_retryable,
        log_error_details,
        safe_call
    )
except ImportError:
    # Error handler utilities not available, but this is optional
    pass

try:
    from .transfer import (
        RetryOpts,
        get_shard_configs,
        calculate_price,
        delay,
        get_split_num,
        segment_range,
        wait_for_receipt,
        tx_with_gas_adjustment,
        submit_with_gas_adjustment
    )
except ImportError:
    # Transfer utilities not available - may be due to missing dependencies
    pass



"""
File uploader for 0G Storage.

Ported from official TypeScript SDK:
node_modules/@0glabs/0g-ts-sdk/lib.commonjs/transfer/Uploader.js

CRITICAL: Must EXACTLY match TypeScript SDK behavior.
This is the MAIN upload orchestration file - 405 lines in TS SDK.
"""
from typing import List, Dict, Any, Optional, Tuple
import base64
import time
from web3 import Web3
from eth_account.signers.local import LocalAccount

try:
    from ..contracts.flow import FlowContract
    from ..core.storage_node import StorageNode
    from ..core.file import ZgFile
    from ..core.merkle import MerkleTree
    from ..config import (
        DEFAULT_CHUNK_SIZE,
        DEFAULT_SEGMENT_SIZE,
        DEFAULT_SEGMENT_MAX_CHUNKS
    )
    from ..utils.transfer import (
        get_shard_configs,
        calculate_price,
        delay,
        segment_range
    )
    from ..core.node_selector import check_replica
    from ..core.encrypted_file import (
        EncryptedFile,
        new_ecies_encrypted_file,
        new_symmetric_encrypted_file,
    )
except ImportError:
    from contracts.flow import FlowContract
    from core.storage_node import StorageNode
    from core.file import ZgFile
    from core.merkle import MerkleTree
    from config import (
        DEFAULT_CHUNK_SIZE,
        DEFAULT_SEGMENT_SIZE,
        DEFAULT_SEGMENT_MAX_CHUNKS
    )
    from utils.transfer import (
        get_shard_configs,
        calculate_price,
        delay,
        segment_range
    )
    from core.node_selector import check_replica
    from core.encrypted_file import (
        EncryptedFile,
        new_ecies_encrypted_file,
        new_symmetric_encrypted_file,
    )


class Uploader:
    """
    File uploader for 0G Storage.

    Ported from Uploader.js (lines 1-407).

    Handles the complete upload workflow:
    1. Submit transaction to Flow contract
    2. Split file into uploadable segments
    3. Upload segments to storage nodes in parallel
    4. Handle retries and errors
    """

    def __init__(
        self,
        nodes: List[StorageNode],
        provider_rpc: str,
        flow: FlowContract,
        gas_price: int = 0,
        gas_limit: int = 0
    ):
        """
        Initialize uploader.

        TS SDK lines 15-21.

        Args:
            nodes: List of storage node clients
            provider_rpc: Blockchain RPC URL
            flow: Flow contract instance
            gas_price: Gas price (0 for auto)
            gas_limit: Gas limit (0 for auto)
        """
        # TS line 16
        self.nodes = nodes

        # TS line 17
        self.web3 = Web3(Web3.HTTPProvider(provider_rpc))

        # TS line 18
        self.flow = flow

        # TS line 19-20
        self.gas_price = gas_price
        self.gas_limit = gas_limit

    def upload_file(
        self,
        file: ZgFile,
        opts: Dict[str, Any],
        retry_opts: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, str], Optional[Exception]]:
        """
        Upload file to storage network.

        TS SDK lines 22-85.

        Args:
            file: File to upload
            opts: Upload options
            retry_opts: Retry options

        Returns:
            Tuple of (result_dict, error)
        """
        # TS line 63-64: wrap with encryption before computing the merkle tree.
        encryption = opts.get('encryption')
        if encryption:
            file = self._wrap_encryption(file, encryption)

        # TS line 23-28
        tree, err = file.merkle_tree()
        if err is not None or tree is None or tree.root_hash() is None:
            return (
                {'txHash': '', 'rootHash': ''},
                Exception('Failed to create Merkle tree')
            )

        # TS line 30
        root_hash = tree.root_hash()

        # TS line 31
        print(
            f"Data prepared to upload root={root_hash} " +
            f"size={file.size()} " +
            f"numSegments={file.num_segments()} " +
            f"numChunks={file.num_chunks()}"
        )

        # TS line 32-33
        tx_seq = None
        receipt = None

        # TS line 34
        info = self.find_existing_file_info(root_hash)

        # TS line 35
        if not opts.get('skipTx', False) or info is None:
            # TS line 36-41
            # Get submitter address from options or account
            account = opts.get('account')
            submitter = opts.get('submitter', '')
            if not submitter and account:
                submitter = account.address
            if not submitter:
                submitter = "0x0000000000000000000000000000000000000000"
            
            submission, err = file.create_submission(opts.get('tags', b'\x00'), submitter)
            if err is not None or submission is None:
                return (
                    {'txHash': '', 'rootHash': root_hash},
                    Exception('Failed to create submission')
                )

            # TS line 43-46
            tx_receipt, tx_err = self.submit_transaction(submission, opts, retry_opts)
            if tx_err is not None:
                return ({'txHash': '', 'rootHash': root_hash}, tx_err)

            # TS line 47
            receipt = tx_receipt

            # TS line 48
            tx_hash_display = receipt['transactionHash'].hex()
            if not tx_hash_display.startswith('0x'):
                tx_hash_display = f"0x{tx_hash_display}"
            print(f"Transaction hash: {tx_hash_display}")

            # TS line 49-55
            tx_seqs = self.flow.process_logs(receipt)
            if len(tx_seqs) == 0:
                return (
                    {'txHash': '', 'rootHash': root_hash},
                    Exception('Failed to get txSeqs')
                )

            # TS line 56
            print(f"Transaction sequence number: {tx_seqs[0]}")

            # TS line 57
            tx_seq = tx_seqs[0]

            # TS line 58
            info = self.wait_for_log_entry(tx_seq, False)

        # TS line 60
        # Ensure transaction hash has 0x prefix for blockchain explorers
        if receipt:
            tx_hash_hex = receipt['transactionHash'].hex()
            tx_hash = tx_hash_hex if tx_hash_hex.startswith('0x') else f"0x{tx_hash_hex}"
        else:
            tx_hash = ''

        # TS line 61-63
        if info is None:
            return ({'txHash': tx_hash, 'rootHash': root_hash}, Exception('Failed to get log entry'))

        # TS line 64-70
        tasks = self.split_tasks(info, tree, opts)
        if tasks is None:
            return (
                {'txHash': tx_hash, 'rootHash': root_hash},
                Exception('Failed to get upload tasks')
            )

        # TS line 71-73
        if len(tasks) == 0:
            return ({'txHash': tx_hash, 'rootHash': root_hash}, None)

        # TS line 74
        print(f"Processing tasks in parallel with {len(tasks)} tasks...")

        # TS line 75
        results = self.process_tasks_in_parallel(file, tree, tasks, retry_opts)

        # TS line 77-81
        # Check for errors, but don't fail if nodes propagate via network
        has_errors = False
        for i in range(len(results)):
            if isinstance(results[i], Exception):
                has_errors = True
                print(f'⚠️  Task {i} had error (non-fatal): {str(results[i])}')

        # TS line 82
        if has_errors:
            print('⚠️  Some direct uploads failed, but file may still propagate via network')
        else:
            print('✅ All tasks processed successfully')

        # TS line 83
        # Wait for finality if required - this ensures network propagation
        if opts.get('finalityRequired', False):
            self.wait_for_log_entry(info['tx']['seq'], True)

        # TS line 84
        # Return success even if direct uploads failed, as network propagates files
        return ({'txHash': tx_hash, 'rootHash': root_hash}, None)

    def splitable_upload(
        self,
        file: ZgFile,
        opts: Dict[str, Any],
        retry_opts: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, list], Optional[Exception]]:
        """
        Upload file with automatic splitting into fragments if it exceeds fragmentSize.
        
        TS SDK Uploader.ts lines 172-231.
        
        Args:
            file: File to upload
            opts: Upload options (should include 'fragmentSize' for max fragment size)
            retry_opts: Retry options
            
        Returns:
            Tuple of ({txHashes: [], rootHashes: []}, error)
        """
        from ..utils.file_utils import next_pow2
        from ..config import DEFAULT_CHUNK_SIZE, DEFAULT_BATCH_SIZE
        
        # Get fragment size from options, default to 4GB
        fragment_size = opts.get('fragmentSize', 4 * 1024 * 1024 * 1024)
        
        # Ensure fragment size is at least chunk size
        if fragment_size < DEFAULT_CHUNK_SIZE:
            fragment_size = DEFAULT_CHUNK_SIZE
        
        # Align size of fragment to power of 2
        fragment_size = next_pow2(fragment_size)
        
        tx_hashes: list = []
        root_hashes: list = []
        
        if file.size() <= fragment_size:
            # Single file upload
            result, err = self.upload_file(file, opts, retry_opts)
            if err is not None:
                return ({'txHashes': tx_hashes, 'rootHashes': root_hashes}, err)
            tx_hashes.append(result['txHash'])
            root_hashes.append(result['rootHash'])
        else:
            # Split and batch upload
            fragments = file.split(fragment_size)
            print(f"Split origin file into {len(fragments)} fragments, {fragment_size} bytes each.")
            
            batch_size = opts.get('batchSize', DEFAULT_BATCH_SIZE)
            
            for l in range(0, len(fragments), batch_size):
                r = min(l + batch_size, len(fragments))
                print(f"Batch uploading fragments {l} to {r}...")
                
                # Process fragments sequentially to maintain order
                for i in range(l, r):
                    result, err = self.upload_file(fragments[i], opts, retry_opts)
                    if err is not None:
                        return ({'txHashes': tx_hashes, 'rootHashes': root_hashes}, err)
                    tx_hashes.append(result['txHash'])
                    root_hashes.append(result['rootHash'])
        
        return ({'txHashes': tx_hashes, 'rootHashes': root_hashes}, None)

    def submit_transaction(
        self,
        submission: Dict[str, Any],
        opts: Dict[str, Any],
        retry_opts: Optional[Dict[str, Any]]
    ) -> Tuple[Optional[Dict], Optional[Exception]]:
        """
        Submit transaction to Flow contract.

        TS SDK lines 86-123.

        Args:
            submission: Submission structure
            opts: Transaction options
            retry_opts: Retry options

        Returns:
            Tuple of (receipt, error)
        """
        # TS line 87-89 - Market contract for pricing
        fee = 0
        if 'fee' in opts and opts['fee'] > 0:
            fee = opts['fee']
        else:
            # Calculate storage fee from market contract
            try:
                # Get market contract address from flow contract
                market_addr = self.flow.contract.functions.market().call()

                # Get market contract and call pricePerSector()
                from .market import get_market_contract
                market_contract = get_market_contract(market_addr, self.web3)
                price_per_sector = market_contract.functions.pricePerSector().call()

                # Calculate fee: sectors * pricePerSector
                # sectors = sum of (1 << node.height) for each node
                # Note: submission has new structure with 'data' wrapper
                sectors = 0
                nodes = submission.get('data', submission).get('nodes', submission.get('nodes', []))
                for node in nodes:
                    sectors += 1 << int(node['height'])

                fee = sectors * price_per_sector
                print(f"Calculated storage fee from market contract: {fee}")
            except Exception as e:
                # Fallback: if market contract fails, use zero fee
                # The transaction may still succeed depending on contract state
                print(f"Warning: Failed to calculate storage fee ({type(e).__name__}): {e}")
                fee = 0

        # TS line 97-113
        account = opts.get('account')
        if account is None:
            return (None, Exception('Account required for transaction'))

        tx_params = {
            'from': account.address,
            'value': fee,
        }

        if 'nonce' in opts:
            tx_params['nonce'] = opts['nonce']

        if self.gas_price > 0:
            tx_params['gasPrice'] = self.gas_price
        else:
            suggested_gas = self.web3.eth.gas_price
            if suggested_gas is None:
                return (
                    None,
                    Exception('Failed to get suggested gas price, set your own gas price')
                )
            tx_params['gasPrice'] = suggested_gas

        # TS line 114-116
        if self.gas_limit > 0:
            tx_params['gas'] = self.gas_limit

        # TS line 117
        print(f"Submitting transaction with storage fee: {fee}")

        # TS line 118-122
        try:
            receipt = self.flow.submit(submission, account, value=fee)
            return (receipt, None)
        except Exception as e:
            return (None, Exception(f'Failed to submit transaction: {str(e)}'))

    def find_existing_file_info(self, root_hash: str) -> Optional[Dict[str, Any]]:
        """
        Find existing file info from storage nodes.

        TS SDK lines 124-140.

        Args:
            root_hash: File root hash

        Returns:
            File info or None
        """
        # TS line 125
        print('Attempting to find existing file info by root hash...')

        # TS line 127-138
        for client in self.nodes:
            try:
                info = client.get_file_info(root_hash, False)
                if info is not None:
                    print(f"Found existing file info: {info}")
                    return info
            except Exception as e:
                print(f"Failed to get file info from node: {client.url}")

        # TS line 139
        return None

    def wait_for_log_entry(
        self,
        tx_seq: int,
        finality_required: bool
    ) -> Optional[Dict[str, Any]]:
        """
        Wait for log entry to be available on storage nodes.

        TS SDK lines 190-220.

        Args:
            tx_seq: Transaction sequence number
            finality_required: Whether to wait for finality

        Returns:
            File info or None
        """
        # TS line 191
        print('Wait for log entry on storage node')

        # TS line 192
        info = None

        # TS line 193
        while True:
            # TS line 194
            delay(1)  # 1 second

            # TS line 195
            ok = True

            # TS line 196
            for client in self.nodes:
                # TS line 197
                info = client.get_file_info_by_tx_seq(tx_seq)

                # TS line 198-208
                if info is None:
                    log_msg = 'Log entry is unavailable yet'
                    status = client.get_status()
                    if status is not None:
                        log_sync_height = status['logSyncHeight']
                        log_msg = f"Log entry is unavailable yet, zgsNodeSyncHeight={log_sync_height}"
                    print(log_msg)
                    ok = False
                    break

                # TS line 209-213
                if finality_required and not info['finalized']:
                    print(f"Log entry is available, but not finalized yet, {client} {info}")
                    ok = False
                    break

            # TS line 215-217
            if ok:
                break

        # TS line 219
        return info

    def process_tasks_in_parallel(
        self,
        file: ZgFile,
        tree: MerkleTree,
        tasks: List[Dict[str, Any]],
        retry_opts: Optional[Dict[str, Any]]
    ) -> List[Any]:
        """
        Process upload tasks in parallel.

        TS SDK lines 222-225.

        NOTE: Python doesn't have built-in async like JS.
        For MVP, we'll process sequentially. Can add asyncio later.

        Args:
            file: File to upload
            tree: Merkle tree
            tasks: Upload tasks
            retry_opts: Retry options

        Returns:
            List of results
        """
        # TS line 223-224 - Promise.all in TypeScript
        # For Python MVP, process sequentially
        results = []
        for task in tasks:
            result = self.upload_task(file, tree, task, retry_opts)
            results.append(result)
        return results

    def next_segment_index(self, config: Dict[str, int], start_index: int) -> int:
        """
        Calculate next segment index for shard.

        TS SDK lines 226-234.

        Args:
            config: Shard configuration
            start_index: Starting index

        Returns:
            Next segment index
        """
        # TS line 227-229
        if config['numShard'] < 2:
            return start_index

        # TS line 230-233
        return (
            ((start_index + config['numShard'] - 1 - config['shardId']) //
             config['numShard']) *
            config['numShard'] +
            config['shardId']
        )

    def split_tasks(
        self,
        info: Dict[str, Any],
        tree: MerkleTree,
        opts: Dict[str, Any]
    ) -> Optional[List[Dict[str, Any]]]:
        """
        Split upload into tasks for each shard.

        TS SDK lines 235-285.

        Args:
            info: File info from log entry
            tree: Merkle tree
            opts: Upload options

        Returns:
            List of upload tasks or None
        """
        # TS line 236-240
        shard_configs = get_shard_configs(self.nodes)
        if shard_configs is None:
            print('Failed to get shard configs')
            return None

        # TS line 241-244
        if not check_replica(shard_configs, opts.get('expectedReplica', 1)):
            print('Not enough replicas')
            return None

        # TS line 245
        tx_seq = info['tx']['seq']

        # TS line 246
        start_segment_index, end_segment_index = segment_range(
            info['tx']['startEntryIndex'],
            info['tx']['size']
        )

        # TS line 247
        upload_tasks = []

        # Calculate expected number of segments
        num_segments = end_segment_index - start_segment_index + 1

        # Check if file is already fully uploaded across ALL required shards
        # In a sharded network, we need to verify EACH shard has its assigned segments
        all_shards_complete = True
        for client_index in range(len(shard_configs)):
            c_info = self.nodes[client_index].get_file_info(tree.root_hash(), True)

            # Check if this shard has uploaded all its segments
            if c_info is not None and c_info.get('finalized', False):
                uploaded_segments = c_info.get('uploadedSegNum', 0)

                # Calculate how many segments this specific shard should have
                shard_config = shard_configs[client_index]
                seg_index = self.next_segment_index(shard_config, start_segment_index)
                expected_segs_for_shard = 0
                while seg_index <= end_segment_index:
                    expected_segs_for_shard += 1
                    seg_index += shard_config['numShard']

                if uploaded_segments < expected_segs_for_shard:
                    all_shards_complete = False
                    print(f"⚠️  Shard {shard_config['shardId']} incomplete: {uploaded_segments}/{expected_segs_for_shard} segments")
            else:
                # Shard doesn't have the file at all
                all_shards_complete = False

        if all_shards_complete:
            print(f"✅ File fully uploaded across all shards - Skipping upload")
            return []

        # TS line 248
        for client_index in range(len(shard_configs)):
            # TS line 249
            shard_config = shard_configs[client_index]

            # TS line 250-254
            # Skip this node if it already has all segments uploaded
            c_info = self.nodes[client_index].get_file_info(tree.root_hash(), True)
            if c_info is not None and c_info.get('finalized', False):
                uploaded_segments = c_info.get('uploadedSegNum', 0)
                if uploaded_segments >= num_segments:
                    print(f"Node {self.nodes[client_index].url} already has all segments, skipping")
                    continue
                # If finalized but missing segments, continue to upload

            # TS line 255
            tasks = []

            # TS line 256
            seg_index = self.next_segment_index(shard_config, start_segment_index)

            # TS line 257
            while seg_index <= end_segment_index:
                # TS line 258-264
                tasks.append({
                    'clientIndex': client_index,
                    'taskSize': opts.get('taskSize', 1),
                    'segIndex': seg_index - start_segment_index,
                    'numShard': shard_config['numShard'],
                    'txSeq': tx_seq,
                })
                # TS line 265
                seg_index += shard_config['numShard'] * opts.get('taskSize', 1)

            # TS line 267-269
            if len(tasks) > 0:
                upload_tasks.append(tasks)

        # TS line 271-273
        if len(upload_tasks) == 0:
            return []

        # TS line 274
        print(f"Tasks created: {upload_tasks}")

        # TS line 275
        tasks = []

        # TS line 276-283
        if len(upload_tasks) > 0:
            upload_tasks.sort(key=lambda a: len(a))
            for task_index in range(len(upload_tasks[0])):
                for i in range(len(upload_tasks)):
                    if task_index < len(upload_tasks[i]):
                        tasks.append(upload_tasks[i][task_index])

        # TS line 284
        return tasks

    def get_segment(
        self,
        file: ZgFile,
        tree: MerkleTree,
        seg_index: int
    ) -> Tuple[bool, Optional[Dict[str, Any]], Optional[Exception]]:
        """
        Get segment data with proof.

        TS SDK lines 286-314.

        Args:
            file: File object
            tree: Merkle tree
            seg_index: Segment index

        Returns:
            Tuple of (all_data_uploaded, segment_with_proof, error)
        """
        # TS line 287
        num_chunks = file.num_chunks()

        # TS line 288-291
        start_seg_index = seg_index * DEFAULT_SEGMENT_MAX_CHUNKS
        if start_seg_index >= num_chunks:
            return (True, None, None)

        # TS line 292
        iter = file.iterate_with_offset_and_batch(
            seg_index * DEFAULT_SEGMENT_SIZE,
            DEFAULT_SEGMENT_SIZE,
            True
        )

        # TS line 293-296
        ok, err = iter.next()
        if not ok:
            return (False, None, err)

        # TS line 297
        segment = iter.current()

        # TS line 298
        proof = tree.proof_at(seg_index)

        # TS line 299
        start_index = seg_index * DEFAULT_SEGMENT_MAX_CHUNKS

        # TS line 300
        all_data_uploaded = False

        # TS line 301-305
        if start_index + len(segment) // DEFAULT_CHUNK_SIZE >= num_chunks:
            expected_len = DEFAULT_CHUNK_SIZE * (num_chunks - start_index)
            segment = segment[0:expected_len]
            all_data_uploaded = True

        # TS line 306-312
        seg_with_proof = {
            'root': tree.root_hash(),
            'data': base64.b64encode(segment).decode('ascii'),
            'index': seg_index,
            'proof': {
                'lemma': proof.lemma,
                'path': proof.path,
            },
            'fileSize': file.size(),
        }

        # TS line 313
        return (all_data_uploaded, seg_with_proof, None)

    def upload_task(
        self,
        file: ZgFile,
        tree: MerkleTree,
        upload_task: Dict[str, Any],
        retry_opts: Optional[Dict[str, Any]]
    ) -> Any:
        """
        Upload a single task (batch of segments).

        TS SDK lines 315-381.

        Args:
            file: File object
            tree: Merkle tree
            upload_task: Task definition
            retry_opts: Retry options

        Returns:
            Result or Error
        """
        # TS line 316
        seg_index = upload_task['segIndex']

        # TS line 317
        segments = []

        # TS line 318
        for i in range(upload_task['taskSize']):
            # TS line 319-322
            all_data_uploaded, seg_with_proof, err = self.get_segment(file, tree, seg_index)
            if err is not None:
                return err

            # TS line 323-325
            if seg_with_proof is not None:
                segments.append(seg_with_proof)

            # TS line 326-328
            if all_data_uploaded:
                break

            # TS line 329
            seg_index += upload_task['numShard']

        # TS line 332
        max_retries = retry_opts.get('TooManyDataRetries', 3) if retry_opts else 3

        # TS line 333-334
        attempt = 0
        last_error = None

        # TS line 335
        while attempt < max_retries:
            try:
                # TS line 337
                node_url = self.nodes[upload_task['clientIndex']].url
                print(f"Uploading {len(segments)} segment(s) to {node_url}, attempt {attempt + 1}/{max_retries}...")
                # Debug: print first segment structure (without data)
                if len(segments) > 0:
                    seg_debug = {k: v for k, v in segments[0].items() if k != 'data'}
                    print(f"  Segment structure: {seg_debug}")
                res = self.nodes[upload_task['clientIndex']].upload_segments_by_tx_seq(
                    segments,
                    upload_task['txSeq']
                )

                # TS line 338-340
                if res is None:
                    raise Exception(
                        f"Node {self.nodes[upload_task['clientIndex']].url} " +
                        "returned null for upload segments"
                    )

                # TS line 341
                return res

            except Exception as error:
                # TS line 344
                last_error = error
                node_url = self.nodes[upload_task['clientIndex']].url

                # TS line 347-350
                if self.is_already_uploaded_error(error):
                    print(f"Segments already uploaded and finalized on node {node_url}")
                    return 0  # Success

                # Handle "Invalid params: root" error by retrying without root field
                if "Invalid params: root" in str(error):
                    print(f"Node {node_url} rejects 'root' field, retrying without it...")
                    try:
                        segments_without_root = []
                        for seg in segments:
                            seg_copy = seg.copy()
                            seg_copy.pop('root', None)
                            segments_without_root.append(seg_copy)

                        res = self.nodes[upload_task['clientIndex']].upload_segments_by_tx_seq(
                            segments_without_root,
                            upload_task['txSeq']
                        )
                        if res is None:
                            raise Exception(
                                f"Node {node_url} returned null for upload segments"
                            )
                        return res
                    except Exception as fallback_error:
                        last_error = fallback_error
                        print(f"Fallback without 'root' also failed: {str(fallback_error)}")
                        # Continue to normal error handling below

                # TS line 352
                if self.is_retryable_error(error):
                    # TS line 353
                    if attempt < max_retries - 1:
                        # TS line 354
                        wait_time = (retry_opts.get('Interval', 3) if retry_opts else 3) * (attempt + 1)

                        # TS line 355
                        error_type = self.get_error_type(error)

                        # TS line 356
                        print(
                            f"{error_type} on attempt {attempt + 1}/{max_retries}. " +
                            f"Retrying in {wait_time}s..."
                        )

                        # TS line 357
                        delay(wait_time)

                        # TS line 358-359
                        attempt += 1
                        continue
                    else:
                        # TS line 361-366
                        error_message = str(error)
                        print(f"Max retries ({max_retries}) reached for error: {error_message}")
                        return Exception(f"Failed after {max_retries} attempts: {error_message}")
                else:
                    # TS line 369-373
                    error_message = str(error)
                    print(f"Non-retryable error encountered: {error_message}")
                    return last_error

        # TS line 377-380
        final_error = last_error or Exception(
            f"Upload failed after {max_retries} attempts to node " +
            f"{self.nodes[upload_task['clientIndex']].url}"
        )
        print(f"Upload task failed completely: {str(final_error)}")
        return final_error

    def is_already_uploaded_error(self, error: Exception) -> bool:
        """
        Check if error indicates file already uploaded.

        TS SDK lines 382-387.
        """
        error_str = str(error).lower()
        return (
            'already uploaded and finalized' in error_str or
            ('invalid params' in error_str and 'already uploaded' in error_str)
        )

    def is_retryable_error(self, error: Exception) -> bool:
        """
        Check if error is retryable.

        TS SDK lines 388-393.
        """
        error_str = str(error)
        return (
            'too many data writing' in error_str or
            'returned null for upload segments' in error_str
        )

    def get_error_type(self, error: Exception) -> str:
        """
        Get error type description.

        TS SDK lines 394-404.
        """
        error_str = str(error)
        if 'too many data writing' in error_str:
            return '"too many data writing" error'
        elif 'returned null' in error_str:
            return 'null response error'
        else:
            return 'retryable error'

    def _wrap_encryption(self, file, encryption: Dict[str, Any]):
        """
        Wrap ``file`` with an EncryptedFile per the ``encryption`` option.

        Mirrors TS ``Uploader.wrapEncryption``. Accepted shapes:
            {"type": "aes256", "key": bytes}
            {"type": "ecies", "recipient_pub_key": bytes | hex_str}
        """
        enc_type = encryption.get('type')
        if enc_type == 'aes256':
            key = encryption.get('key')
            if key is None:
                raise ValueError("aes256 encryption requires 'key'")
            return new_symmetric_encrypted_file(file, key)
        if enc_type == 'ecies':
            recipient = (
                encryption.get('recipient_pub_key')
                or encryption.get('recipientPubKey')
            )
            if recipient is None:
                raise ValueError("ecies encryption requires 'recipient_pub_key'")
            return new_ecies_encrypted_file(file, recipient)
        raise ValueError(f"unsupported encryption type: {enc_type!r}")



"""
Cryptographic utilities for merkle tree and hashing.

Ported from official TypeScript SDK which uses:
- @ethersproject/keccak256 for hashing
- @ethersproject/bytes for hex operations
"""
from typing import List
from Crypto.Hash import keccak


def keccak256_hash(data: bytes) -> str:
    """
    Compute Keccak256 hash of data.

    Returns hex string with 0x prefix (matching ethers.js behavior).
    """
    k = keccak.new(digest_bits=256)
    k.update(data)
    return '0x' + k.hexdigest()


def keccak256_hash_bytes(data: bytes) -> bytes:
    """
    Compute Keccak256 hash of data.

    Returns raw bytes (for internal use).
    """
    k = keccak.new(digest_bits=256)
    k.update(data)
    return k.digest()


def hex_concat(hex_strings: List[str]) -> bytes:
    """
    Concatenate multiple hex strings into bytes.

    Matches @ethersproject/bytes hexConcat behavior.
    """
    result = b''
    for hex_str in hex_strings:
        result += hex_to_bytes(hex_str)
    return result


def keccak256_hash_combine(*hashes: str) -> str:
    """
    Combine multiple hashes by concatenating then hashing.

    Matches TS implementation:
    function keccak256Hash(...hashes) {
        return keccak256(hexConcat(hashes));
    }
    """
    combined = hex_concat(list(hashes))
    return keccak256_hash(combined)


def bytes_to_hex(data: bytes) -> str:
    """Convert bytes to hex string with 0x prefix."""
    return '0x' + data.hex()


def hex_to_bytes(hex_str: str) -> bytes:
    """Convert hex string to bytes."""
    if hex_str.startswith('0x'):
        hex_str = hex_str[2:]
    return bytes.fromhex(hex_str)



"""
Stream data builder for KV operations.

Ported from official TypeScript SDK:
src.ts/kv/builder.ts
"""
from typing import Dict, Optional
from .types import AccessControl, StreamData, StreamRead, StreamWrite
from .constants import MAX_KEY_SIZE, MAX_SET_SIZE, STREAM_DOMAIN

# Type aliases
Hash = str


class StreamDataBuilder:
    """
    Builder for constructing StreamData objects.
    
    TS SDK lines 9-155.
    """
    
    def __init__(self, version: int):
        """
        Initialize builder.
        
        TS SDK lines 16-22.
        
        Args:
            version: Stream data version
        """
        self.version = version
        self.stream_ids: Dict[Hash, bool] = {}
        self.controls: list[AccessControl] = []
        self.reads: Dict[Hash, Dict[str, bool]] = {}
        self.writes: Dict[Hash, Dict[str, bytes]] = {}

    def _hex_to_bytes(self, hex_str: str) -> bytes:
        """Convert hex string to bytes."""
        if hex_str.startswith("0x"):
            hex_str = hex_str[2:]
        return bytes.fromhex(hex_str)

    def build(self, sorted_: bool = False) -> StreamData:
        """
        Build StreamData from accumulated operations.
        
        TS SDK lines 32-109.
        
        Args:
            sorted_: Whether to sort reads and writes by stream_id and key
            
        Returns:
            Constructed StreamData object
        """
        data = StreamData(version=self.version)
        
        # controls
        data.controls = self._build_access_control()
        
        # reads
        data.reads = []
        for stream_id, keys in self.reads.items():
            for k in keys.keys():
                key = self._hex_to_bytes(k)
                if len(key) > MAX_KEY_SIZE:
                    raise ValueError("Key too large")
                if len(key) == 0:
                    raise ValueError("Key is empty")
                data.reads.append(StreamRead(
                    stream_id=stream_id,
                    key=key
                ))
                
                if len(data.reads) > MAX_SET_SIZE:
                    raise ValueError("Size too large")
        
        # writes
        data.writes = []
        for stream_id, keys in self.writes.items():
            for k, d in keys.items():
                key = self._hex_to_bytes(k)
                if len(key) > MAX_KEY_SIZE:
                    raise ValueError("Key too large")
                if len(key) == 0:
                    raise ValueError("Key is empty")
                data.writes.append(StreamWrite(
                    stream_id=stream_id,
                    key=key,
                    data=d
                ))
                
                if len(data.writes) > MAX_SET_SIZE:
                    raise ValueError("Size too large")
        
        if sorted_:
            # Sort reads by stream_id, then by key
            data.reads.sort(key=lambda x: (x.stream_id, x.key.hex()))
            # Sort writes by stream_id, then by key
            data.writes.sort(key=lambda x: (x.stream_id, x.key.hex()))
        
        return data

    def set(self, stream_id: str, key: bytes, data: bytes) -> None:
        """
        Add a write operation.
        
        TS SDK lines 111-121.
        
        Args:
            stream_id: Stream ID (hex string)
            key: Key bytes
            data: Value bytes
        """
        self.add_stream_id(stream_id)
        
        if stream_id not in self.writes:
            self.writes[stream_id] = {}
        
        self.writes[stream_id][key.hex()] = data

    def add_stream_id(self, stream_id: Hash) -> None:
        """
        Register a stream ID.
        
        TS SDK lines 123-125.
        
        Args:
            stream_id: Stream ID (hex string)
        """
        self.stream_ids[stream_id] = True

    def build_tags(self, sorted_: bool = False) -> bytes:
        """
        Build tags for upload.
        
        TS SDK lines 127-135.
        
        Args:
            sorted_: Whether to sort stream IDs
            
        Returns:
            Encoded tags bytes
        """
        ids = list(self.stream_ids.keys())
        
        if sorted_:
            ids.sort()
        
        return self._create_tags(ids)

    def _create_tags(self, stream_ids: list[Hash]) -> bytes:
        """
        Create tags from stream IDs.
        
        TS SDK lines 137-147.
        
        Args:
            stream_ids: List of stream IDs
            
        Returns:
            Encoded tags
        """
        result = bytearray((1 + len(stream_ids)) * 32)
        
        # Set stream domain
        result[0:32] = STREAM_DOMAIN
        
        # Set each stream ID
        for index, id_ in enumerate(stream_ids):
            id_bytes = self._hex_to_bytes(id_)
            result[32 * (index + 1):32 * (index + 2)] = id_bytes
        
        return bytes(result)

    def _build_access_control(self) -> list[AccessControl]:
        """
        Build access controls.
        
        TS SDK lines 149-154.
        
        Returns:
            List of access controls
        """
        if len(self.controls) > MAX_SET_SIZE:
            raise ValueError("Size too large")
        return self.controls



"""
KV Batcher for batching KV operations and uploading.

Ported from official TypeScript SDK:
src.ts/kv/batcher.ts
"""
from typing import Optional, Dict, Any, List, Tuple

from .builder import StreamDataBuilder

try:
    from ..core.storage_node import StorageNode
    from ..core.file import ZgFile
    from ..core.uploader import Uploader
    from ..contracts.flow import FlowContract
except ImportError:
    from core.storage_node import StorageNode
    from core.file import ZgFile
    from core.uploader import Uploader
    from contracts.flow import FlowContract


# Default upload options
DEFAULT_UPLOAD_OPTION: Dict[str, Any] = {
    'tags': b'\x00',
    'finalityRequired': True,
    'taskSize': 10,
    'expectedReplica': 1,
    'skipTx': False,
}


class Batcher:
    """
    Batcher for KV operations.
    
    TS SDK lines 11-51.
    
    Accumulates KV write operations and uploads them as a batch.
    """

    def __init__(
        self,
        version: int,
        clients: List[StorageNode],
        flow: FlowContract,
        provider: str
    ):
        """
        Initialize batcher.
        
        TS SDK lines 17-27.
        
        Args:
            version: Stream data version
            clients: List of storage node clients
            flow: Flow contract instance
            provider: Blockchain RPC URL
        """
        self.stream_data_builder = StreamDataBuilder(version)
        self.clients = clients
        self.flow = flow
        self.blockchain_rpc = provider

    def set(self, stream_id: str, key: bytes, data: bytes) -> None:
        """
        Add a write operation to the batch.
        
        Args:
            stream_id: Stream ID (hex string)
            key: Key bytes
            data: Value bytes
        """
        self.stream_data_builder.set(stream_id, key, data)

    def exec(
        self,
        opts: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, str], Optional[Exception]]:
        """
        Execute the batched operations by uploading to storage.
        
        TS SDK lines 29-50.
        
        Args:
            opts: Upload options (optional)
            
        Returns:
            Tuple of ({txHash, rootHash}, error)
        """
        # Build stream data
        stream_data = self.stream_data_builder.build()
        encoded = stream_data.encode()
        
        # Create in-memory file from encoded data
        data = ZgFile.from_bytes(encoded)
        
        # Create uploader
        uploader = Uploader(
            self.clients,
            self.blockchain_rpc,
            self.flow
        )
        
        # Set up options
        if opts is None:
            opts = DEFAULT_UPLOAD_OPTION.copy()
        
        # Add tags from builder
        opts['tags'] = self.stream_data_builder.build_tags()
        
        # Upload the file
        return uploader.upload_file(data, opts)



"""
Best-effort post-download decryption helpers.

Direct port of ``src.ts/indexer/decryption.ts``. Both helpers are
intentionally non-throwing for decryption-related reasons: if the file
isn't actually encrypted, the embedded ephemeral pubkey isn't on-curve,
or the wrong key material was supplied, they return the raw bytes
unchanged with ``decrypted=False`` so the caller can fall back.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Union

try:
    from .encryption import (
        ECIES_VERSION,
        EncryptionHeader,
        KeyLike,
        decrypt_file,
        decrypt_fragment_data,
        parse_encryption_header,
        resolve_decryption_key,
    )
except ImportError:  # pragma: no cover - test runner fallback
    from core.encryption import (
        ECIES_VERSION,
        EncryptionHeader,
        KeyLike,
        decrypt_file,
        decrypt_fragment_data,
        parse_encryption_header,
        resolve_decryption_key,
    )

from cryptography.hazmat.primitives.asymmetric import ec


_SECP256K1 = ec.SECP256K1()


@dataclass
class TryDecryptResult:
    bytes: bytes
    decrypted: bool


def _ephemeral_on_curve(ephemeral_pub: bytes) -> bool:
    try:
        ec.EllipticCurvePublicKey.from_encoded_point(_SECP256K1, ephemeral_pub)
        return True
    except Exception:
        return False


def try_decrypt(
    encrypted: bytes,
    symmetric_key: Optional[bytes] = None,
    private_key: Optional[KeyLike] = None,
) -> TryDecryptResult:
    """
    Best-effort decrypt. Returns ``(bytes, decrypted)``. Never throws for
    decryption-related reasons — any failure returns the raw bytes with
    ``decrypted=False``.
    """
    if len(encrypted) < 1:
        return TryDecryptResult(bytes=encrypted, decrypted=False)

    try:
        header = parse_encryption_header(encrypted)
    except Exception:
        return TryDecryptResult(bytes=encrypted, decrypted=False)

    # AES-CTR is unauthenticated; without the on-curve check a plaintext
    # file that happens to start with 0x02 would "decrypt" to garbage.
    if header.version == ECIES_VERSION and not _ephemeral_on_curve(
        header.ephemeral_pub
    ):
        return TryDecryptResult(bytes=encrypted, decrypted=False)

    try:
        aes_key = resolve_decryption_key(symmetric_key, private_key, header)
    except Exception:
        return TryDecryptResult(bytes=encrypted, decrypted=False)

    try:
        plaintext = decrypt_file(aes_key, encrypted)
    except Exception:
        return TryDecryptResult(bytes=encrypted, decrypted=False)
    return TryDecryptResult(bytes=plaintext, decrypted=True)


def try_decrypt_fragments(
    fragments: List[bytes],
    symmetric_key: Optional[bytes] = None,
    private_key: Optional[KeyLike] = None,
) -> Optional[List[bytes]]:
    """
    Best-effort multi-fragment decrypt. Mirrors 0g-storage-client's
    ``indexer/client.go::downloadEncryptedFragments``: parse the header from
    fragment 0, resolve the AES key, then decrypt each subsequent fragment at
    the correct CTR offset so mid-stream fragments decrypt identically to a
    full-file decrypt.

    Returns a list of plaintext fragments, or ``None`` on any failure so the
    caller can fall back to the raw concatenation.
    """
    if len(fragments) == 0:
        return []

    try:
        header = parse_encryption_header(fragments[0])
    except Exception:
        return None

    if header.version == ECIES_VERSION and not _ephemeral_on_curve(
        header.ephemeral_pub
    ):
        return None

    try:
        aes_key = resolve_decryption_key(symmetric_key, private_key, header)
    except Exception:
        return None

    plaintexts: List[bytes] = []
    cumulative_offset = 0
    for i, fragment in enumerate(fragments):
        try:
            result = decrypt_fragment_data(
                aes_key,
                header,
                fragment,
                i == 0,
                cumulative_offset,
            )
        except Exception:
            return None
        plaintexts.append(result.plaintext)
        cumulative_offset = result.new_offset
    return plaintexts



"""
Input validation utilities.
"""
import os


def validate_file_path(path: str) -> bool:
    """Check if file path exists and is readable."""
    return os.path.isfile(path) and os.access(path, os.R_OK)


def validate_root_hash(root_hash: str) -> bool:
    """Check if root hash is valid hex string."""
    if not root_hash:
        return False

    # Remove 0x prefix if present
    if root_hash.startswith('0x'):
        root_hash = root_hash[2:]

    # Check if valid hex and reasonable length
    try:
        int(root_hash, 16)
        return len(root_hash) == 64  # 32 bytes = 64 hex chars
    except ValueError:
        return False


def validate_replicas(replicas: int) -> bool:
    """Check if replica count is valid."""
    return replicas >= 1



"""
TS <-> Python interop tests.

Runs the real `@0gfoundation/0g-ts-sdk` encryption code (from
`.cache/0g_ts_sdk/`) via Node and verifies that:

1. Python's v1 encryption produces byte-identical output to TS for fixed
   key + nonce + plaintext (the canonical reference implementation).
2. Python can decrypt a v1 file produced by TS.
3. Python can decrypt a v2 (ECIES) file produced by TS (random ephemeral,
   so compare plaintext, not ciphertext).
4. TS can decrypt v1 and v2 files produced by Python.

If any test fails, Python and TS will not interoperate at the wire
layer — files encrypted in one cannot be decrypted in the other.

Tests are skipped unless `node` and the cached TS SDK with installed
dependencies are present, so normal CI runs without Node are unaffected.
"""

import json
import shutil
import subprocess
import sys
from pathlib import Path

import pytest

sys.path.insert(0, str(Path(__file__).parent.parent))

from core.encryption import (
    SYMMETRIC_VERSION,
    EncryptionHeader,
    crypt_at,
    decrypt_file,
    derive_ecies_decrypt_key,
)
from core.encrypted_file import new_ecies_encrypted_file


INTEROP_DIR = Path(__file__).parent / "_interop"
TS_SDK_ROOT = Path(__file__).parent.parent.parent / ".cache" / "0g_ts_sdk"
TSX = TS_SDK_ROOT / "node_modules" / ".bin" / "tsx"

_skip_reason = None
if shutil.which("node") is None:
    _skip_reason = "node not on PATH"
elif not TSX.exists():
    _skip_reason = f"tsx not installed at {TSX} (run `npm install` in .cache/0g_ts_sdk)"
elif not TS_SDK_ROOT.joinpath("src.ts/common/encryption.ts").exists():
    _skip_reason = "TS SDK source not present in .cache/0g_ts_sdk"

pytestmark = pytest.mark.skipif(
    _skip_reason is not None, reason=_skip_reason or ""
)


def _run_ts(script_name: str, payload: dict) -> bytes:
    """Run a TS helper with the manifest as stdin JSON, return stdout hex bytes."""
    script = INTEROP_DIR / script_name
    result = subprocess.run(
        [str(TSX), str(script)],
        input=json.dumps(payload).encode("utf-8"),
        capture_output=True,
        check=False,
    )
    if result.returncode != 0:
        raise RuntimeError(
            f"TS helper {script_name} failed:\nstdout: {result.stdout!r}\nstderr: {result.stderr!r}"
        )
    return bytes.fromhex(result.stdout.decode("utf-8").strip())


# ── Determinism: identical key+nonce+plaintext → identical ciphertext ──────


def test_v1_python_matches_ts_byte_for_byte():
    """
    Strongest possible interop check: with a fixed AES key and nonce, both
    sides must produce *byte-identical* encrypted files. Any divergence
    here means a wire incompatibility.
    """
    key = bytes([0x42] * 32)
    nonce = bytes([0x13] * 16)
    plaintext = b"interop check: python and TS must agree on every byte"

    # Python side
    py_header = EncryptionHeader(SYMMETRIC_VERSION, nonce)
    py_body = crypt_at(key, nonce, 0, plaintext)
    py_encrypted = py_header.to_bytes() + py_body

    # TS side
    ts_encrypted = _run_ts(
        "ts_encrypt_v1.mjs",
        {
            "key_hex": key.hex(),
            "nonce_hex": nonce.hex(),
            "plaintext_hex": plaintext.hex(),
        },
    )

    assert py_encrypted == ts_encrypted, (
        f"v1 ciphertext divergence:\n"
        f"  python: {py_encrypted.hex()}\n"
        f"  ts:     {ts_encrypted.hex()}"
    )


# ── Round-trip: TS encrypts, Python decrypts (v1) ──────────────────────────


def test_python_decrypts_ts_v1_output():
    key = bytes([0x77] * 32)
    nonce = bytes(range(16))
    plaintext = b"a TS-encrypted v1 file that Python must read"

    ts_encrypted = _run_ts(
        "ts_encrypt_v1.mjs",
        {
            "key_hex": key.hex(),
            "nonce_hex": nonce.hex(),
            "plaintext_hex": plaintext.hex(),
        },
    )

    recovered = decrypt_file(key, ts_encrypted)
    assert recovered == plaintext


# ── Round-trip: TS encrypts (random ephemeral), Python decrypts (v2) ──────


def test_python_decrypts_ts_v2_output():
    """
    ECIES uses a random ephemeral key per encryption so ciphertext bytes
    will differ run-to-run, but Python must still recover the plaintext
    using the recipient's static private key.
    """
    recipient_priv = bytes([0x11] * 32)
    # priv = 0x11*32 -> compressed pub. Derive it the same way Python does.
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import ec

    priv_obj = ec.derive_private_key(
        int.from_bytes(recipient_priv, "big"), ec.SECP256K1()
    )
    recipient_pub = priv_obj.public_key().public_bytes(
        encoding=serialization.Encoding.X962,
        format=serialization.PublicFormat.CompressedPoint,
    )

    plaintext = b"v2 ECIES handshake across the language boundary"

    ts_encrypted = _run_ts(
        "ts_encrypt_v2.mjs",
        {
            "recipient_pub_hex": recipient_pub.hex(),
            "plaintext_hex": plaintext.hex(),
        },
    )

    # Python parses the header, derives the AES key via ECDH, decrypts.
    from core.encryption import parse_encryption_header

    header = parse_encryption_header(ts_encrypted)
    aes_key = derive_ecies_decrypt_key(recipient_priv, header.ephemeral_pub)
    recovered = decrypt_file(aes_key, ts_encrypted)
    assert recovered == plaintext


# ── Round-trip: Python encrypts, TS decrypts (v1) ──────────────────────────


class _MemFile:
    def __init__(self, data):
        self._data = data

    def size(self):
        return len(self._data)

    def read_from_file(self, start, end):
        chunk = self._data[start:end]
        return len(chunk), chunk

    def create_fragment(self, *a):
        raise NotImplementedError

    def iterate_with_offset_and_batch(self, *a):
        raise NotImplementedError


def test_ts_decrypts_python_v1_output():
    from core.encrypted_file import new_symmetric_encrypted_file

    key = bytes([0x99] * 32)
    plaintext = b"python wrote this; TS must read it"

    ef = new_symmetric_encrypted_file(_MemFile(plaintext), key)
    _, py_encrypted = ef.read_from_file(0, ef.size())

    recovered = _run_ts(
        "ts_decrypt.mjs",
        {
            "encrypted_hex": py_encrypted.hex(),
            "symmetric_key_hex": key.hex(),
        },
    )
    assert recovered == plaintext


# ── Round-trip: Python encrypts, TS decrypts (v2) ──────────────────────────


def test_ts_decrypts_python_v2_output():
    recipient_priv_hex = "01" * 32
    recipient_pub_hex = (
        "031b84c5567b126440995d3ed5aaba0565d71e1834604819ff9c17f5e9d5dd078f"
    )
    plaintext = b"python ECIES wrote this; TS recovers it"

    ef = new_ecies_encrypted_file(_MemFile(plaintext), recipient_pub_hex)
    _, py_encrypted = ef.read_from_file(0, ef.size())

    recovered = _run_ts(
        "ts_decrypt.mjs",
        {
            "encrypted_hex": py_encrypted.hex(),
            "private_key_hex": "0x" + recipient_priv_hex,
        },
    )
    assert recovered == plaintext



"""
Transfer utilities.

Ported from official TypeScript SDK:
- node_modules/@0glabs/0g-ts-sdk/lib.commonjs/transfer/utils.js
- node_modules/@0glabs/0g-ts-sdk/lib.commonjs/utils.js

CRITICAL: Must EXACTLY match TypeScript SDK behavior.
"""
from typing import List, Dict, Any, Optional, Tuple, Callable, TYPE_CHECKING
from dataclasses import dataclass
import time

from web3 import Web3
from web3.contract import Contract
from web3.types import TxReceipt
from eth_account.signers.local import LocalAccount

# Use TYPE_CHECKING to avoid circular imports
if TYPE_CHECKING:
    from ..core.storage_node import StorageNode


@dataclass
class RetryOpts:
    """
    Options for transaction retry with gas adjustment.
    
    TS SDK types.ts RetryOpts.
    """
    retries: int = 10
    interval: int = 5  # seconds
    max_gas_price: int = 0  # 0 means no limit


def _get_storage_node_class():
    """Lazy import to avoid circular dependency."""
    try:
        from ..core.storage_node import StorageNode
        return StorageNode
    except ImportError:
        from core.storage_node import StorageNode
        return StorageNode


def _get_is_valid_config():
    """Lazy import to avoid circular dependency."""
    try:
        from ..core.node_selector import is_valid_config
        return is_valid_config
    except ImportError:
        from core.node_selector import is_valid_config
        return is_valid_config


def get_shard_configs(nodes: List["StorageNode"]) -> Optional[List[Dict[str, Any]]]:
    """
    Get shard configurations from storage nodes.

    TS SDK transfer/utils.js lines 6-16.

    Args:
        nodes: List of storage nodes

    Returns:
        List of shard configs or None if any invalid
    """
    is_valid_config = _get_is_valid_config()
    configs = []
    for c_node in nodes:
        c_config = c_node.get_shard_config()
        if not is_valid_config(c_config):
            return None
        configs.append(c_config)
    return configs


def calculate_price(submission: Dict[str, Any], price_per_sector: int) -> int:
    """
    Calculate storage price for submission.

    TS SDK transfer/utils.js lines 17-23.

    Args:
        submission: Submission structure
        price_per_sector: Price per sector

    Returns:
        Total price
    """
    sectors = 0
    for node in submission['nodes']:
        sectors += 1 << int(node['height'])

    return sectors * price_per_sector


def delay(seconds: float) -> None:
    """
    Delay execution.

    TS SDK utils.js line 22.

    Args:
        seconds: Seconds to delay
    """
    time.sleep(seconds)


def get_split_num(total: int, unit: int) -> int:
    """
    Calculate number of splits.

    TS SDK utils.js lines 38-40.

    Args:
        total: Total size
        unit: Unit size

    Returns:
        Number of splits
    """
    return (total - 1) // unit + 1


def segment_range(start_chunk_index: int, file_size: int) -> tuple:
    """
    Calculate segment range for file.

    TS SDK utils.js lines 49-58.

    Args:
        start_chunk_index: Starting chunk index
        file_size: File size in bytes

    Returns:
        Tuple of (start_segment_index, end_segment_index)
    """
    try:
        from ..config import DEFAULT_CHUNK_SIZE, DEFAULT_SEGMENT_MAX_CHUNKS
    except ImportError:
        from config import DEFAULT_CHUNK_SIZE, DEFAULT_SEGMENT_MAX_CHUNKS

    # TS line 51
    total_chunks = get_split_num(file_size, DEFAULT_CHUNK_SIZE)

    # TS line 53
    start_segment_index = start_chunk_index // DEFAULT_SEGMENT_MAX_CHUNKS

    # TS line 55-56
    end_chunk_index = start_chunk_index + total_chunks - 1
    end_segment_index = end_chunk_index // DEFAULT_SEGMENT_MAX_CHUNKS

    # TS line 57
    return (start_segment_index, end_segment_index)


def wait_for_receipt(
    web3: Web3,
    tx_hash: str,
    opts: Optional[RetryOpts] = None
) -> Optional[TxReceipt]:
    """
    Wait for transaction receipt with retries.
    
    TS SDK utils.ts lines 125-156.
    
    Args:
        web3: Web3 instance
        tx_hash: Transaction hash
        opts: Retry options
        
    Returns:
        Transaction receipt or None if timeout
    """
    if opts is None:
        opts = RetryOpts(retries=10, interval=5, max_gas_price=0)
    
    if opts.retries == 0:
        opts.retries = 10
    if opts.interval == 0:
        opts.interval = 5
    
    n_tries = 0
    
    while n_tries < opts.retries:
        try:
            receipt = web3.eth.get_transaction_receipt(tx_hash)
            
            if receipt is not None and receipt['status'] == 1:
                return receipt
        except Exception:
            pass
        
        delay(opts.interval)
        n_tries += 1
    
    return None


def tx_with_gas_adjustment(
    web3: Web3,
    contract: Contract,
    account: LocalAccount,
    method: str,
    params: List[Any],
    tx_opts: Dict[str, Any],
    retry_opts: Optional[RetryOpts] = None
) -> Tuple[Optional[TxReceipt], Optional[Exception]]:
    """
    Execute transaction with automatic gas price adjustment on timeout.
    
    TS SDK utils.ts lines 72-123.
    
    If a transaction times out, this function will retry with a 10% higher gas price
    until the max_gas_price is reached.
    
    Args:
        web3: Web3 instance
        contract: Contract instance
        account: Account to sign transaction
        method: Contract method name
        params: Method parameters
        tx_opts: Transaction options (must include 'value' and 'gasPrice')
        retry_opts: Retry options with max_gas_price
        
    Returns:
        Tuple of (receipt, error) - one will be None
    """
    current_gas_price = tx_opts.get('gasPrice', web3.eth.gas_price)
    max_gas_price = current_gas_price
    
    if retry_opts is not None and retry_opts.max_gas_price > 0:
        max_gas_price = retry_opts.max_gas_price
    
    err: Optional[Exception] = None
    
    while current_gas_price <= max_gas_price:
        print(f"Sending transaction with gas price {current_gas_price}")
        tx_opts['gasPrice'] = current_gas_price
        
        try:
            # Get the contract function
            contract_fn = getattr(contract.functions, method)
            
            # Build transaction
            tx_params = {
                'from': account.address,
                'value': tx_opts.get('value', 0),
                'nonce': web3.eth.get_transaction_count(account.address),
                'gasPrice': current_gas_price,
            }
            
            # Add gas limit if specified
            if 'gasLimit' in tx_opts or 'gas' in tx_opts:
                tx_params['gas'] = tx_opts.get('gasLimit', tx_opts.get('gas'))
            else:
                # Estimate gas
                try:
                    tx_params['gas'] = contract_fn(*params).estimate_gas(tx_params)
                except Exception:
                    tx_params['gas'] = 500000  # Fallback
            
            # Build and sign transaction
            tx = contract_fn(*params).build_transaction(tx_params)
            signed_tx = account.sign_transaction(tx)
            
            # Send transaction
            tx_hash = web3.eth.send_raw_transaction(signed_tx.raw_transaction)
            
            # Wait for transaction to be mined
            receipt = web3.eth.wait_for_transaction_receipt(tx_hash, timeout=120)
            
            if receipt is None:
                raise Exception('Send transaction timeout')
            
            # Wait for receipt with retries
            final_receipt = wait_for_receipt(
                web3,
                tx_hash.hex() if hasattr(tx_hash, 'hex') else tx_hash,
                retry_opts
            )
            
            if final_receipt is None:
                raise Exception('Get transaction receipt timeout')
            
            return (final_receipt, None)
            
        except Exception as e:
            err = e
            error_msg = str(e).lower()
            
            # Check if it's a timeout error
            if 'timeout' in error_msg:
                print(
                    f"Failed to send transaction with gas price {current_gas_price}, "
                    f"with error {e}, retrying with higher gas price"
                )
                # Increase gas price by 10% (multiply by 11/10)
                current_gas_price = (11 * current_gas_price) // 10
                
                delay(1)  # Wait 1 second before retry
            else:
                # Non-timeout error, return immediately
                return (None, err)
    
    return (None, err)


def submit_with_gas_adjustment(
    web3: Web3,
    flow_contract: Contract,
    account: LocalAccount,
    submission: Dict[str, Any],
    fee: int,
    gas_price: int,
    gas_limit: Optional[int] = None,
    retry_opts: Optional[RetryOpts] = None
) -> Tuple[Optional[TxReceipt], Optional[Exception]]:
    """
    Submit to Flow contract with gas auto-adjustment.
    
    Convenience wrapper for tx_with_gas_adjustment for Flow.submit().
    
    Args:
        web3: Web3 instance
        flow_contract: Flow contract instance
        account: Account to sign transaction
        submission: Submission data structure
        fee: Storage fee in wei
        gas_price: Initial gas price
        gas_limit: Optional gas limit
        retry_opts: Retry options with max_gas_price
        
    Returns:
        Tuple of (receipt, error) - one will be None
    """
    tx_opts = {
        'value': fee,
        'gasPrice': gas_price,
    }
    
    if gas_limit is not None and gas_limit > 0:
        tx_opts['gasLimit'] = gas_limit
    
    return tx_with_gas_adjustment(
        web3=web3,
        contract=flow_contract,
        account=account,
        method='submit',
        params=[submission],
        tx_opts=tx_opts,
        retry_opts=retry_opts
    )



"""
Tests for client-side encryption primitives.

Mirrors `.cache/0g_ts_sdk/tests/encryption.test.ts` 1:1 and uses the TS
reference vectors verbatim. A passing run here proves Python encryption is
wire-compatible with the TS SDK and (transitively) the Go storage client.
"""

import sys
from pathlib import Path

import pytest

sys.path.insert(0, str(Path(__file__).parent.parent))

from core.encryption import (
    ECIES_HEADER_SIZE,
    ECIES_VERSION,
    EPHEMERAL_PUBKEY_SIZE,
    SYMMETRIC_HEADER_SIZE,
    SYMMETRIC_VERSION,
    EncryptionHeader,
    crypt_at,
    decrypt_file,
    decrypt_fragment_data,
    derive_ecies_decrypt_key,
    derive_ecies_encrypt_key,
    new_ecies_header,
    new_symmetric_header,
    normalize_priv_key,
    normalize_pub_key,
    parse_encryption_header,
    resolve_decryption_key,
)


# ── EncryptionHeader v1 ────────────────────────────────────────────────────


def test_v1_serializes_to_17_bytes_and_round_trips():
    nonce = bytes(range(1, 17))
    h = EncryptionHeader(SYMMETRIC_VERSION, nonce)
    raw = h.to_bytes()

    assert len(raw) == SYMMETRIC_HEADER_SIZE == 17
    assert raw[0] == 0x01
    assert raw[1:17] == nonce

    parsed = parse_encryption_header(raw)
    assert parsed.version == SYMMETRIC_VERSION
    assert parsed.nonce == nonce
    assert parsed.size() == 17


def test_v1_parse_rejects_too_short():
    with pytest.raises(ValueError, match="too short"):
        parse_encryption_header(b"\x01")


def test_parse_rejects_unsupported_version():
    bad = bytes([0xEE]) + b"\x00" * 16
    with pytest.raises(ValueError, match="unsupported"):
        parse_encryption_header(bad)


# ── EncryptionHeader v2 ────────────────────────────────────────────────────


def test_v2_serializes_to_50_bytes_and_round_trips():
    nonce = bytes(0xA0 + i for i in range(16))
    ephemeral_pub = bytes(range(EPHEMERAL_PUBKEY_SIZE))

    h = EncryptionHeader(ECIES_VERSION, nonce, ephemeral_pub)
    raw = h.to_bytes()

    assert len(raw) == ECIES_HEADER_SIZE == 50
    assert raw[0] == 0x02
    assert raw[1 : 1 + EPHEMERAL_PUBKEY_SIZE] == ephemeral_pub
    assert raw[1 + EPHEMERAL_PUBKEY_SIZE : 1 + EPHEMERAL_PUBKEY_SIZE + 16] == nonce

    parsed = parse_encryption_header(raw)
    assert parsed.version == ECIES_VERSION
    assert parsed.ephemeral_pub == ephemeral_pub
    assert parsed.nonce == nonce
    assert parsed.size() == 50


def test_v2_parse_rejects_short_data():
    bad = bytes([0x02]) + b"\x00" * 16
    with pytest.raises(ValueError, match="too short"):
        parse_encryption_header(bad)


# ── crypt_at (AES-256-CTR with offset) ─────────────────────────────────────

KEY = bytes([0x42] * 32)
NONCE = bytes([0x13] * 16)


def test_round_trips_at_offset_zero():
    plain = b"hello world encryption test data"
    cipher = crypt_at(KEY, NONCE, 0, plain)
    assert cipher != plain
    decrypted = crypt_at(KEY, NONCE, 0, cipher)
    assert decrypted == plain


def test_two_halves_match_single_full_encrypt():
    original = bytes(range(100))
    full = crypt_at(KEY, NONCE, 0, original)

    part1 = crypt_at(KEY, NONCE, 0, original[:50])
    part2 = crypt_at(KEY, NONCE, 50, original[50:])

    assert part1 == full[:50]
    assert part2 == full[50:]


def test_sub_range_decryption_with_matching_offset():
    plain = bytes(range(100))
    cipher = crypt_at(KEY, NONCE, 0, plain)

    slice_ = cipher[32:]
    decrypted = crypt_at(KEY, NONCE, 32, slice_)
    assert decrypted == plain[32:]


def test_no_op_on_empty_input():
    assert crypt_at(KEY, NONCE, 0, b"") == b""


def test_non_block_aligned_offsets():
    plain = bytes((i * 3) & 0xFF for i in range(64))
    cipher = crypt_at(KEY, NONCE, 0, plain)

    slice_ = cipher[7:]
    decrypted = crypt_at(KEY, NONCE, 7, slice_)
    assert decrypted == plain[7:]


# ── normalize_pub_key — TS reference vector ────────────────────────────────

# From the TS test suite. priv = 0x01*32 -> compressed pub:
COMPRESSED_HEX = "031b84c5567b126440995d3ed5aaba0565d71e1834604819ff9c17f5e9d5dd078f"
COMPRESSED_BYTES = bytes.fromhex(COMPRESSED_HEX)


def test_normalize_pub_key_passes_compressed_through():
    assert normalize_pub_key(COMPRESSED_BYTES).hex() == COMPRESSED_HEX


def test_normalize_pub_key_accepts_compressed_hex_with_prefix():
    assert normalize_pub_key("0x" + COMPRESSED_HEX).hex() == COMPRESSED_HEX


def test_normalize_pub_key_accepts_compressed_hex_no_prefix():
    assert normalize_pub_key(COMPRESSED_HEX).hex() == COMPRESSED_HEX


def test_normalize_pub_key_rejects_invalid_length():
    with pytest.raises(ValueError):
        normalize_pub_key(b"\x00" * 10)


# ── normalize_priv_key ─────────────────────────────────────────────────────

PRIV_HEX = "01" * 32


def test_normalize_priv_key_passes_bytes_through():
    assert normalize_priv_key(bytes.fromhex(PRIV_HEX)).hex() == PRIV_HEX


def test_normalize_priv_key_accepts_hex_with_prefix():
    assert normalize_priv_key("0x" + PRIV_HEX).hex() == PRIV_HEX


def test_normalize_priv_key_accepts_hex_no_prefix():
    assert normalize_priv_key(PRIV_HEX).hex() == PRIV_HEX


def test_normalize_priv_key_rejects_short_input():
    with pytest.raises(ValueError):
        normalize_priv_key(b"\x00" * 10)
    with pytest.raises(ValueError):
        normalize_priv_key("beef")


def test_normalize_priv_key_rejects_zero_scalar():
    with pytest.raises(ValueError, match="invalid"):
        normalize_priv_key(b"\x00" * 32)


# ── ECIES key derivation ───────────────────────────────────────────────────


def test_ecies_encrypt_and_decrypt_keys_match():
    # Use the TS reference vector for the recipient.
    recipient_priv = bytes.fromhex(PRIV_HEX)
    recipient_pub = COMPRESSED_BYTES

    enc_key, ephemeral_pub = derive_ecies_encrypt_key(recipient_pub)
    dec_key = derive_ecies_decrypt_key(recipient_priv, ephemeral_pub)

    assert len(enc_key) == 32
    assert len(ephemeral_pub) == 33
    assert dec_key == enc_key


def test_ecies_generates_fresh_ephemeral_each_call():
    recipient_pub = COMPRESSED_BYTES
    _, ephemeral_a = derive_ecies_encrypt_key(recipient_pub)
    _, ephemeral_b = derive_ecies_encrypt_key(recipient_pub)
    assert ephemeral_a != ephemeral_b


def test_ecies_wrong_private_key_yields_different_aes_key():
    recipient_priv = bytes.fromhex(PRIV_HEX)
    attacker_priv = bytes.fromhex("02" * 32)
    recipient_pub = COMPRESSED_BYTES

    right_key, ephemeral_pub = derive_ecies_encrypt_key(recipient_pub)
    wrong_key = derive_ecies_decrypt_key(attacker_priv, ephemeral_pub)
    assert wrong_key != right_key


# ── End-to-end: encrypt with one factory, decrypt with the other ───────────


def test_full_file_round_trip_v1():
    header = new_symmetric_header()
    key = bytes([0x77] * 32)
    plaintext = b"the quick brown fox jumps over the lazy dog" * 10

    body = crypt_at(key, header.nonce, 0, plaintext)
    encrypted_file = header.to_bytes() + body

    decrypted = decrypt_file(key, encrypted_file)
    assert decrypted == plaintext


def test_full_file_round_trip_v2():
    recipient_priv = bytes.fromhex(PRIV_HEX)
    recipient_pub = COMPRESSED_BYTES

    header, key = new_ecies_header(recipient_pub)
    plaintext = b"ecies round trip" * 32

    body = crypt_at(key, header.nonce, 0, plaintext)
    encrypted_file = header.to_bytes() + body

    parsed_header = parse_encryption_header(encrypted_file)
    decrypt_key = resolve_decryption_key(None, recipient_priv, parsed_header)
    decrypted = decrypt_file(decrypt_key, encrypted_file)
    assert decrypted == plaintext


def test_decrypt_fragment_data_first_and_subsequent():
    header = new_symmetric_header()
    key = bytes([0x99] * 32)
    plaintext = bytes(range(200))
    full_cipher = header.to_bytes() + crypt_at(key, header.nonce, 0, plaintext)

    # First fragment: header + first 80 bytes of ciphertext.
    split = header.size() + 80
    first_fragment = full_cipher[:split]
    rest = full_cipher[split:]

    r1 = decrypt_fragment_data(key, header, first_fragment, True, 0)
    assert r1.plaintext == plaintext[:80]
    assert r1.new_offset == 80

    r2 = decrypt_fragment_data(key, header, rest, False, r1.new_offset)
    assert r2.plaintext == plaintext[80:]
    assert r2.new_offset == len(plaintext)


def test_resolve_decryption_key_errors():
    sym_header = EncryptionHeader(SYMMETRIC_VERSION, b"\x00" * 16)
    with pytest.raises(ValueError, match="requires a symmetric key"):
        resolve_decryption_key(None, None, sym_header)
    with pytest.raises(ValueError, match="32 bytes"):
        resolve_decryption_key(b"\x00" * 31, None, sym_header)

    ecies_header = EncryptionHeader(
        ECIES_VERSION, b"\x00" * 16, b"\x02" + b"\x01" * 32
    )
    with pytest.raises(ValueError, match="requires a private key"):
        resolve_decryption_key(None, None, ecies_header)



"""Data models for 0G Storage SDK."""

from .transaction import RetryOpts, TransactionOptions
from .node import ShardConfig, ShardedNode, Status, NetworkIdentity, NetworkProtocolVersion, IpLocation, ShardedNodes
from .file import (
    MerkleNode,
    Segment,
    FileProof,
    SegmentWithProof,
    Transaction,
    FileInfo,
    Metadata,
    Value,
    KeyValue,
    FlowProof
)

__all__ = [
    # Transaction models
    "RetryOpts",
    "TransactionOptions",

    # Node models
    "ShardConfig",
    "ShardedNode",
    "Status",
    "NetworkIdentity",
    "NetworkProtocolVersion",
    "IpLocation",
    "ShardedNodes",

    # File models
    "MerkleNode",
    "Segment",
    "FileProof",
    "SegmentWithProof",
    "Transaction",
    "FileInfo",
    "Metadata",
    "Value",
    "KeyValue",
    "FlowProof",
]
